@@ -5,7 +5,7 @@ from contextlib import ExitStack
 from functools import partial
 import random
 from pathlib import Path
-from itertools import groupby
+from itertools import chain, groupby
 
 import numpy as np
 from sklearn.utils.class_weight import compute_class_weight
@@ -40,6 +40,7 @@ class DataSilo:
         self,
         processor,
         batch_size,
+        eval_batch_size=None,
         distributed=False,
         automatic_loading=True,
         max_multiprocessing_chunksize=2000,
@@ -50,9 +51,12 @@ class DataSilo:
         """
         :param processor: A dataset specific Processor object which will turn input (file or dict) into a Pytorch Dataset.
         :type processor: Processor
-        :param batch_size: The size of batch that should be returned by the DataLoaders.
+        :param batch_size: The size of batch that should be returned by the DataLoader for the training set.
         :type batch_size: int
-        :param distributed: Set to True if the program is running in a distributed setting.
+        :param eval_batch_size: The size of batch that should be returned by the DataLoaders for the dev and test set.
+        :type eval_batch_size: int
+        :param distributed: Set to True if you are running in a distributed evn, e.g. using DistributedDataParallel.
+                            The DataSilo will init the DataLoader with a DistributedSampler() to distribute batches.
         :type distributed: bool
         :param automatic_loading: Set to False, if you don't want to automatically load data at initialization.
         :type automatic_loading: bool
@@ -61,7 +65,7 @@ class DataSilo:
             values are rather large that might cause memory issues.
         :type max_multiprocessing_chunksize: int
         :param max_processes: the maximum number of processes to spawn in the multiprocessing.Pool used in DataSilo.
-                              It can be set to 1 to disable the use of multiprocessing ot make debugging easier.
+                              It can be set to 1 to disable the use of multiprocessing or make debugging easier.
         :type max_processes: int
         :param caching: save the processed datasets on disk to save time/compute if the same train data is used to run
                         multiple experiments. Each cache has a checksum based on the train_filename of the Processor
@@ -80,6 +84,10 @@ class DataSilo:
         self.caching = caching
         self.cache_path = cache_path
         self.tensor_names = None
+        if eval_batch_size is None:
+            self.eval_batch_size = batch_size
+        else:
+            self.eval_batch_size = eval_batch_size
 
         if len(self.processor.tasks) == 0:
             raise Exception("No task initialized. Try initializing the processor with a metric and a label list. "
@@ -121,8 +129,8 @@ class DataSilo:
         """
         dicts = [d[1] for d in chunk]
         indices = [x[0] for x in chunk]
-        dataset = processor.dataset_from_dicts(dicts=dicts, indices=indices)
-        return dataset
+        dataset, tensor_names, problematic_sample_ids = processor.dataset_from_dicts(dicts=dicts, indices=indices)
+        return dataset, tensor_names, problematic_sample_ids
 
     def _get_dataset(self, filename, dicts=None):
         if not filename and not dicts:
@@ -162,21 +170,27 @@ class DataSilo:
             else:
                 logger.info(
                     f"Multiprocessing disabled, using a single worker to convert {num_dicts}"
-                    f"dictionaries to pytorch datasets."
+                    f" dictionaries to pytorch datasets."
                 )
 
                 results = map(partial(self._dataset_from_chunk, processor=self.processor), grouper(dicts, num_dicts))
 
             datasets = []
+            problematic_ids_all = set()
 
             desc = f"Preprocessing Dataset"
             if filename:
                 desc += f" {filename}"
             with tqdm(total=len(dicts), unit=' Dicts', desc=desc) as pbar:
-                for dataset, tensor_names in results:
+                for dataset, tensor_names, problematic_samples in results:
                     datasets.append(dataset)
                     # update progress bar (last step can have less dicts than actual chunk_size)
                     pbar.update(min(multiprocessing_chunk_size, pbar.total-pbar.n))
+                    problematic_ids_all.update(problematic_samples)
+
+            self.processor.log_problematic(problematic_ids_all)
+            # _dataset_from_chunk can return a None in cases where downsampling has occurred
+            datasets = [d for d in datasets if d]
             concat_datasets = ConcatDataset(datasets)
             return concat_datasets, tensor_names
 
@@ -191,9 +205,12 @@ class DataSilo:
         :param test_dicts: (Optional) dicts containing examples for test.
         :return: None
         """
+
         logger.info("\nLoading data into the data silo ..."
                     "{}".format(TRACTOR_SMALL))
         # train data
+        logger.info("LOADING TRAIN DATA")
+        logger.info("==================")
         if train_dicts:
             # either from supplied dicts
             logger.info("Loading train set from supplied dicts ")
@@ -208,6 +225,9 @@ class DataSilo:
             self.data["train"] = None
 
         # dev data
+        logger.info("")
+        logger.info("LOADING DEV DATA")
+        logger.info("=================")
         if dev_dicts:
             # either from supplied dicts
             logger.info("Loading train set from supplied dicts ")
@@ -225,6 +245,9 @@ class DataSilo:
             logger.info("No dev set is being loaded")
             self.data["dev"] = None
 
+        logger.info("")
+        logger.info("LOADING TEST DATA")
+        logger.info("=================")
         # test data
         if test_dicts:
             # either from supplied dicts
@@ -316,7 +339,7 @@ class DataSilo:
     def _initialize_data_loaders(self):
         """ Initializing train, dev and test data loaders for the already loaded datasets """
 
-        if self.processor.train_filename:
+        if self.data["train"] is not None:
             if self.distributed:
                 sampler_train = DistributedSampler(self.data["train"])
             else:
@@ -335,17 +358,17 @@ class DataSilo:
             data_loader_dev = NamedDataLoader(
                 dataset=self.data["dev"],
                 sampler=SequentialSampler(self.data["dev"]),
-                batch_size=self.batch_size,
+                batch_size=self.eval_batch_size,
                 tensor_names=self.tensor_names,
             )
         else:
             data_loader_dev = None
 
-        if self.processor.test_filename:
+        if self.data["test"] is not None:
             data_loader_test = NamedDataLoader(
                 dataset=self.data["test"],
                 sampler=SequentialSampler(self.data["test"]),
-                batch_size=self.batch_size,
+                batch_size=self.eval_batch_size,
                 tensor_names=self.tensor_names,
             )
         else:
@@ -403,11 +426,22 @@ class DataSilo:
 
     def _calculate_statistics(self):
         """ Calculate and log simple summary statistics of the datasets """
+        logger.info("")
+        logger.info("DATASETS SUMMARY")
+        logger.info("================")
 
         self.counts = {}
 
         if self.data["train"]:
             self.counts["train"] = len(self.data["train"])
+            if "input_ids" in self.tensor_names:
+                clipped, ave_len, seq_lens, max_seq_len = self._calc_length_stats_single_encoder()
+            elif "query_input_ids" in self.tensor_names and "passage_input_ids" in self.tensor_names:
+                clipped, ave_len, seq_lens, max_seq_len = self._calc_length_stats_biencoder()
+            else:
+                logger.warning(f"Could not compute length statistics because 'input_ids' or 'query_input_ids' and 'passage_input_ids' are missing.")
+                clipped = -1
+                ave_len = -1
         else:
             self.counts["train"] = 0
 
@@ -421,29 +455,29 @@ class DataSilo:
         else:
             self.counts["test"] = 0
 
-        seq_lens = []
-        if self.data["train"]:
-            for dataset in self.data["train"].datasets:
-                train_input_numpy = dataset[:][0].numpy()
-                seq_lens.extend(np.sum(train_input_numpy != self.processor.tokenizer.pad_token_id, axis=1))
-            max_seq_len = dataset[:][0].shape[1]
-
-        self.clipped = np.mean(np.array(seq_lens) == max_seq_len) if seq_lens else 0
-        self.ave_len = np.mean(seq_lens) if seq_lens else 0
 
         logger.info("Examples in train: {}".format(self.counts["train"]))
         logger.info("Examples in dev  : {}".format(self.counts["dev"]))
         logger.info("Examples in test : {}".format(self.counts["test"]))
         logger.info("")
         if self.data["train"]:
-            logger.info("Longest sequence length observed after clipping:     {}".format(max(seq_lens)))
-            logger.info("Average sequence length after clipping: {}".format(self.ave_len))
-            logger.info("Proportion clipped:      {}".format(self.clipped))
-            if self.clipped > 0.5:
-                logger.info("[Farmer's Tip] {}% of your samples got cut down to {} tokens. "
-                            "Consider increasing max_seq_len. "
-                            "This will lead to higher memory consumption but is likely to "
-                            "improve your model performance".format(round(self.clipped * 100, 1), max_seq_len))
+            if "input_ids" in self.tensor_names:
+                logger.info("Longest sequence length observed after clipping:     {}".format(max(seq_lens)))
+                logger.info("Average sequence length after clipping: {}".format(ave_len))
+                logger.info("Proportion clipped:      {}".format(clipped))
+                if clipped > 0.5:
+                    logger.info("[Farmer's Tip] {}% of your samples got cut down to {} tokens. "
+                                "Consider increasing max_seq_len. "
+                                "This will lead to higher memory consumption but is likely to "
+                                "improve your model performance".format(round(clipped * 100, 1), max_seq_len))
+            elif "query_input_ids" in self.tensor_names and "passage_input_ids" in self.tensor_names:
+                logger.info("Longest query length observed after clipping: {}   - for max_query_len: {}".format(max(seq_lens[0]),max_seq_len[0]))
+                logger.info("Average query length after clipping:          {}".format(ave_len[0]))
+                logger.info("Proportion queries clipped:                   {}".format(clipped[0]))
+                logger.info("")
+                logger.info("Longest passage length observed after clipping: {}   - for max_passage_len: {}".format(max(seq_lens[1]),max_seq_len[1]))
+                logger.info("Average passage length after clipping:          {}".format(ave_len[1]))
+                logger.info("Proportion passages clipped:                    {}".format(clipped[1]))
 
         MlLogger.log_params(
             {
@@ -451,10 +485,42 @@ class DataSilo:
                 "n_samples_dev": self.counts["dev"],
                 "n_samples_test": self.counts["test"],
                 "batch_size": self.batch_size,
-                "ave_seq_len": self.ave_len,
-                "clipped": self.clipped,
+                "ave_seq_len": ave_len,
+                "clipped": clipped,
             }
         )
+        
+    def _calc_length_stats_single_encoder(self):
+        seq_lens = []
+        for dataset in self.data["train"].datasets:
+            train_input_numpy = dataset[:][self.tensor_names.index("input_ids")].numpy()
+            seq_lens.extend(np.sum(train_input_numpy != self.processor.tokenizer.pad_token_id, axis=1))
+        max_seq_len = dataset[:][self.tensor_names.index("input_ids")].shape[1]
+        clipped = np.mean(np.array(seq_lens) == max_seq_len) if seq_lens else 0
+        ave_len = np.mean(seq_lens) if seq_lens else 0
+        return clipped, ave_len, seq_lens, max_seq_len
+
+    def _calc_length_stats_biencoder(self):
+        seq_lens = [[], []]
+        for dataset in self.data["train"].datasets:
+            query_input_numpy = dataset[:][self.tensor_names.index("query_input_ids")].numpy()
+            num_passages = dataset[:][self.tensor_names.index("passage_input_ids")].shape[1]
+            bs = dataset[:][self.tensor_names.index("passage_input_ids")].shape[0]
+            passage_input_numpy = dataset[:][self.tensor_names.index("passage_input_ids")].numpy().reshape((bs,-1), order='C')
+            qlen = np.sum(query_input_numpy != self.processor.query_tokenizer.pad_token_id, axis=1)
+            plen = np.sum(passage_input_numpy != self.processor.passage_tokenizer.pad_token_id, axis=1) / num_passages
+            seq_lens[0].extend(qlen)
+            seq_lens[1].extend(plen)
+        q_max_seq_len = dataset[:][self.tensor_names.index("query_input_ids")].shape[1]
+        p_max_seq_len = dataset[:][self.tensor_names.index("passage_input_ids")].shape[2]
+        clipped_q = np.mean(np.array(seq_lens[0]) == q_max_seq_len) if seq_lens[0] else 0
+        ave_len_q = np.mean(seq_lens[0]) if seq_lens[0] else 0
+        clipped_p = np.mean(np.array(seq_lens[1]) == p_max_seq_len) if seq_lens[1] else 0
+        ave_len_p = np.mean(seq_lens[1]) if seq_lens[1] else 0
+        clipped = [clipped_q, clipped_p]
+        ave_len = [ave_len_q, ave_len_p]
+        max_seq_len = [q_max_seq_len, p_max_seq_len]
+        return clipped, ave_len, seq_lens, max_seq_len
 
     def calculate_class_weights(self, task_name, source="train"):
         """ For imbalanced datasets, we can calculate class weights that can be used later in the
@@ -476,10 +542,17 @@ class DataSilo:
         else:
             raise Exception("source argument expects one of [\"train\", \"all\"]")
         for dataset in datasets:
-            if dataset is not None:
+            if "multilabel" in self.processor.tasks[task_name]["task_type"]:
+                for x in dataset:
+                    observed_labels += [label_list[label_id] for label_id in (x[tensor_idx] == 1).nonzero()]
+            else:
                 observed_labels += [label_list[x[tensor_idx].item()] for x in dataset]
+
         #TODO scale e.g. via logarithm to avoid crazy spikes for rare classes
-        class_weights = list(compute_class_weight("balanced", np.asarray(label_list), observed_labels))
+        class_weights = compute_class_weight("balanced", classes=np.asarray(label_list), y=observed_labels)
+
+        # conversion necessary to have class weights of same type as model weights
+        class_weights = class_weights.astype(np.float32)
         return class_weights
 
     def get_data_loader(self, dataset_name):
@@ -511,7 +584,7 @@ class StreamingDataSilo:
     multiple workers that utilize the available CPU cores and ensure enough pre-computed batches.
     """
 
-    def __init__(self, processor, batch_size, dataloader_workers=8):
+    def __init__(self, processor, batch_size, distributed=False, dataloader_workers=8):
         """
         :param processor: A dataset specific Processor object which will turn input file into a Pytorch Dataset.
         :type processor: Processor
@@ -524,6 +597,7 @@ class StreamingDataSilo:
         self.processor = processor
         self.batch_size = batch_size
         self.dataloader_workers = dataloader_workers
+        self.distributed = distributed
 
     def get_data_loader(self, dataset_name):
         """
@@ -568,12 +642,19 @@ class StreamingDataSilo:
         #  Since the batching is now handled within _StreamingDataSet, we disable the batching on DataLoader side
         #  by initializing the data loader with batch_size as 1.
 
+        if isinstance(filename, Path) and filename.is_dir():
+            filepath = filename
+        else:
+            filepath = self.processor.data_dir / filename
+
         data_set = _StreamingDataSet(
             processor=self.processor,
-            filepath=self.processor.data_dir / filename,
+            filepath=filepath,
             batch_size=self.batch_size,
             dataloader_workers=self.dataloader_workers,
+            distributed = self.distributed
         )
+
         data_loader = NamedDataLoader(
             dataset=data_set, batch_size=1, num_workers=self.dataloader_workers, pin_memory=True
         )
@@ -581,7 +662,7 @@ class StreamingDataSilo:
 
 
 class _StreamingDataSet(IterableDataset):
-    def __init__(self, processor, filepath, batch_size, dataloader_workers):
+    def __init__(self, processor, filepath, batch_size, dataloader_workers, distributed=False, n_samples=None):
         """
         :param processor: A dataset specific Processor object which will turn input file into a Pytorch Dataset.
         :type processor: Processor
@@ -597,16 +678,39 @@ class _StreamingDataSet(IterableDataset):
         self.processor = processor
         self.filepath = filepath
         self.dataloader_workers = dataloader_workers
+        self.distributed = distributed
 
-        # calculate number of samples for __len__()
-        total_lines = sum(1 for line in open(filepath, encoding="utf-8"))
-        empty_lines = sum(1 if line == "\n" else 0 for line in open(filepath, encoding="utf-8"))
-        self.n_samples = total_lines - (2 * empty_lines)
+        # calculate or estimate number of samples so that the data loader can derive number of training steps
+        if filepath.is_file():
+            files = [filepath]
+        else:
+            files = [file for file in filepath.iterdir()]
 
-        self.file_to_dicts_generator = processor.file_to_dicts(filepath)
+        if n_samples:
+            self.n_samples = n_samples
+        else:
+            try:
+                self.n_samples = self.processor.estimate_n_samples(files[0]) * len(files)
+            except AttributeError:
+                AttributeError(f"Could not estimate n_samples for {self.processor.__class__.__name__} in StreamingDataSilo. "
+                                    f"Make sure that your Processor has `estimate_n_samples()` implemented")
+        logger.info(f"Found data for {self.n_samples} samples")
+        self.shuffle_files(files)
+
+        dicts_from_files = [processor.file_to_dicts(file) for file in files]
+        self.file_to_dicts_generator = chain(*dicts_from_files)
+
+        if self.distributed:
+            self.rank = torch.distributed.get_rank()
+            self.world_size = torch.distributed.get_world_size()
 
     def __len__(self):
-        return self.n_samples
+        if self.distributed:
+            # only a heuristic as we don't necessarily split samples equally across ranks
+            len = self.n_samples // self.world_size
+        else:
+            len = self.n_samples
+        return len
 
     def __iter__(self):
         #  With IterableDataset, the same __iter__ is copied over to the multiple workers of
@@ -621,15 +725,15 @@ class _StreamingDataSet(IterableDataset):
         #  (with n=2) will return, [[dictA, dictB], [dictE, dictF] ...] for worker 1 and
         #  [[dictC, dictD], [dictG, dictH] ...] for worker 2.
 
-        if self.dataloader_workers > 1:
-            worker_info = torch.utils.data.get_worker_info()
-            worker_id = worker_info.id
-            dicts = grouper(
-                self.file_to_dicts_generator, n=10, worker_id=worker_id, total_workers=self.dataloader_workers
-            )
+        worker_info = torch.utils.data.get_worker_info()
+        if self.distributed:
+            worker_id = self.rank * worker_info.num_workers + worker_info.id
+            total_workers = self.world_size * worker_info.num_workers
         else:
-            dicts = grouper(self.file_to_dicts_generator, n=10)
+            worker_id = worker_info.id
+            total_workers = self.dataloader_workers
 
+        dicts = grouper(self.file_to_dicts_generator, n=10, worker_id=worker_id, total_workers=total_workers)
         results = map(self._dataset_from_chunk, dicts)
 
         batch = []
@@ -659,14 +763,24 @@ class _StreamingDataSet(IterableDataset):
             logger.info("Skipping a dict chunk as it contains less than 2 documents ...")
             return None, None
         indices = [x[0] for x in chunk]
-        datasets, tensor_names = self.processor.dataset_from_dicts(dicts=dicts, indices=indices)
+        datasets, tensor_names, _ = self.processor.dataset_from_dicts(dicts=dicts, indices=indices)
         return datasets, tensor_names
 
+    def shuffle_files(self, files, seed=None):
+        if not seed:
+            seed = random.randrange(100)
+        random.seed(seed)
+        random.shuffle(files)
+        return files
+
 
 class DataSiloForCrossVal:
     """
-    For performing cross validation, we really want to combine all the instances from all
-    the sets or just some of the sets, then create a different data silo instance for each fold.
+    Perform cross validation or nested cross validation.
+
+    For performing cross validation or nested cross validation, we really want to combine all the
+    instances from all the sets or just some of the sets, then create a different data silo
+    instance for each fold or nested fold.
     Calling DataSiloForCrossVal.make() creates a list of DataSiloForCrossVal instances - one for each fold.
     """
 
@@ -708,7 +822,7 @@ class DataSiloForCrossVal:
 
     @classmethod
     def make(cls, datasilo, sets=["train", "dev", "test"], n_splits=5, shuffle=True, random_state=None,
-             stratified=True, n_neg_answers_per_question=1):
+             stratified=True, n_neg_answers_per_question=1, n_inner_splits=None):
         """
         Create number of folds data-silo-like objects which can be used for training from the
         original data silo passed on.
@@ -723,16 +837,45 @@ class DataSiloForCrossVal:
         :type shuffle: bool
         :param random_state: random state for shuffling
         :type random_state: int
-        :param stratified: if class stratification should be done
+        :param stratified: If class stratification should be done.
+            It is never done with question answering.
         :type stratified: bool
         :param n_neg_answers_per_question: number of negative answers per question to include for training
         :type n_neg_answers_per_question: int
+        :param n_inner_splits: Number of inner splits of a nested cross validation.
+            Default is ``None`` which means to do a normal (not nested) cross validation.
+            If at least 2 is given a nested cross validation is done. In that case the ``n_splits``
+            parameter is the number of outer splits.
+            The outer cross validation splits the data into a test set and a rest set.
+            The inner cross validation splits the rest data into a train set and a dev set.
+            The advantage of a nested cross validation is that it is doing the inner split
+            not just by random but in a more systematic way. When doing model evaluation
+            this also reduces the variance. This is because you train on more different
+            iterations with more different data constellations.
+        :type n_inner_splits: int
         """
+        # check n_inner_splits param
+        if (n_inner_splits is not None) and (not n_inner_splits >= 2):
+            raise ValueError("'n_inner_splits' must be at least 2!")
 
-        if "question_answering" in datasilo.processor.tasks:
-            return cls._make_question_answering(datasilo, sets, n_splits, shuffle, random_state, n_neg_answers_per_question)
+        if "question_answering" in datasilo.processor.tasks and n_inner_splits is None:
+            return cls._make_question_answering(
+                datasilo, sets, n_splits, shuffle, random_state, n_neg_answers_per_question
+            )
+        elif "question_answering" in datasilo.processor.tasks and n_inner_splits is not None:
+            raise NotImplementedError()
+        elif n_inner_splits is None:
+            return cls._make(
+                datasilo, sets, n_splits, shuffle, random_state, stratified
+            )
+        elif n_inner_splits is not None:
+            return cls._make_nested(
+                datasilo, sets, n_splits, shuffle, random_state, stratified,
+                n_inner_splits
+            )
         else:
-            return cls._make(datasilo, sets, n_splits, shuffle, random_state, stratified)
+            raise RuntimeError("Cross validation can not be done under these conditions!")
+
 
     @classmethod
     def _make_question_answering(cls, datasilo, sets=["train", "dev", "test"], n_splits=5, shuffle=True,
@@ -755,8 +898,11 @@ class DataSiloForCrossVal:
         :param n_neg_answers_per_question: number of negative answers per question to include for training
         :type n_neg_answers_per_question: int
         """
-        assert datasilo.tensor_names[4] == "id", f"Expected tensor 'id' at index 4, found {datasilo.tensor_names[4]}"
-        assert datasilo.tensor_names[7] == "labels", f"Expected tensor 'labels' at index 7, found {datasilo.tensor_names[7]}"
+        assert "id" in datasilo.tensor_names, f"Expected tensor 'id' in tensor names, found {datasilo.tensor_names}"
+        assert "labels" in datasilo.tensor_names, f"Expected tensor 'labels' in tensor names, found {datasilo.tensor_names}"
+
+        id_index = datasilo.tensor_names.index("id")
+        label_index = datasilo.tensor_names.index("labels")
 
         sets_to_concat = []
         for setname in sets:
@@ -765,12 +911,17 @@ class DataSiloForCrossVal:
         all_data = ConcatDataset(sets_to_concat)
 
         documents = []
-        keyfunc = lambda x: x[4][0]
+        keyfunc = lambda x: x[id_index][0]
         all_data = sorted(all_data.datasets, key=keyfunc)
         for key, document in groupby(all_data, key=keyfunc):
             documents.append(list(document))
 
-        xval_split = cls._split_for_qa(documents, n_splits, shuffle, random_state)
+        xval_split = cls._split_for_qa(documents = documents,
+                                       id_index=id_index,
+                                       n_splits=n_splits,
+                                       shuffle=shuffle,
+                                       random_state=random_state,
+                                       )
         silos = []
 
         for train_set, test_set in xval_split:
@@ -789,14 +940,14 @@ class DataSiloForCrossVal:
 
             train_samples = []
             for doc in actual_train_set:
-                keyfunc = lambda x: x[4][1]
+                keyfunc = lambda x: x[id_index][1]
                 doc = sorted(doc, key=keyfunc)
                 for key, question in groupby(doc, key=keyfunc):
                     # add all available answrs to train set
                     sample_list = list(question)
                     neg_answer_idx = []
                     for index, sample in enumerate(sample_list):
-                        if sample[7][0][0] or sample[7][0][1]:
+                        if sample[label_index][0][0] or sample[label_index][0][1]:
                             train_samples.append(sample)
                         else:
                             neg_answer_idx.append(index)
@@ -858,8 +1009,8 @@ class DataSiloForCrossVal:
         return silos
 
     @staticmethod
-    def _split_for_qa(documents, n_splits=5, shuffle=True, random_state=None):
-        keyfunc = lambda x: x[4][1]
+    def _split_for_qa(documents, id_index, n_splits=5, shuffle=True, random_state=None):
+        keyfunc = lambda x: x[id_index][1]
         if shuffle:
             random.shuffle(documents, random_state)
 
@@ -888,3 +1039,45 @@ class DataSiloForCrossVal:
             current_train_set = np.hstack(np.delete(splits, idx, axis=0))
 
             yield current_train_set, current_test_set
+
+    @staticmethod
+    def _make_nested(datasilo, sets=["train", "dev", "test"],
+                     n_splits=5, shuffle=True, random_state=None,
+                     stratified=True, n_inner_splits=5):
+        setstoconcat = [datasilo.data[setname] for setname in sets]
+        ds_all = ConcatDataset(setstoconcat)
+        idxs = list(range(len(ds_all)))
+
+        silos = []
+
+        # outer cross validation where we split all data to test and rest
+        if stratified:
+            # get all the labels for stratification
+            ytensors = [t[3][0] for t in ds_all]
+            y = torch.stack(ytensors)
+            outer_cv = StratifiedKFold(n_splits=n_splits, shuffle=shuffle, random_state=random_state)
+            outer_split = outer_cv.split(idxs, y)
+        else:
+            outer_cv = KFold(n_splits=n_splits, shuffle=shuffle, random_state=random_state)
+            outer_split = outer_cv.split(idxs)
+        for idxs_rest, idxs_test in outer_split:
+
+            # inner cross validation where we split rest data into train and dev
+            if stratified:
+                y_rest = y[idxs_rest]
+                inner_cv = StratifiedKFold(n_splits=n_inner_splits, shuffle=shuffle, random_state=random_state)
+                inner_split = inner_cv.split(idxs_rest, y_rest)
+            else:
+                inner_cv = KFold(n_splits=n_inner_splits, shuffle=shuffle, random_state=random_state)
+                inner_split = inner_cv.split(idxs_rest)
+            for idxs_train_idxs, idxs_dev_idxs in inner_split:
+
+                # split idxs_rest with indexes from inner cross validation
+                idxs_train = idxs_rest[idxs_train_idxs]
+                idxs_dev = idxs_rest[idxs_dev_idxs]
+
+                ds_train = Subset(ds_all, idxs_train)
+                ds_dev = Subset(ds_all, idxs_dev)
+                ds_test = Subset(ds_all, idxs_test)
+                silos.append(DataSiloForCrossVal(datasilo, ds_train, ds_dev, ds_test))
+        return silos
