@@ -5,6 +5,8 @@ import os
 import random
 import tarfile
 import tempfile
+import string
+from contextlib import ExitStack
 from itertools import islice
 from pathlib import Path
 
@@ -21,8 +23,13 @@ logger = logging.getLogger(__name__)
 DOWNSTREAM_TASK_MAP = {
     "gnad": "https://s3.eu-central-1.amazonaws.com/deepset.ai-farm-downstream/gnad.tar.gz",
     "germeval14": "https://s3.eu-central-1.amazonaws.com/deepset.ai-farm-downstream/germeval14.tar.gz",
+
+    # only has train.tsv and test.tsv dataset - no dev.tsv
     "germeval18": "https://s3.eu-central-1.amazonaws.com/deepset.ai-farm-downstream/germeval18.tar.gz",
+
     "squad20": "https://s3.eu-central-1.amazonaws.com/deepset.ai-farm-downstream/squad20.tar.gz",
+    "covidqa": "https://s3.eu-central-1.amazonaws.com/deepset.ai-farm-downstream/covidqa.tar.gz",
+
     "conll03detrain": "https://raw.githubusercontent.com/MaviccPRP/ger_ner_evals/master/corpora/conll2003/deu.train",
     "conll03dedev": "https://raw.githubusercontent.com/MaviccPRP/ger_ner_evals/master/corpora/conll2003/deu.testa", #https://www.clips.uantwerpen.be/conll2003/ner/000README says testa is dev data
     "conll03detest": "https://raw.githubusercontent.com/MaviccPRP/ger_ner_evals/master/corpora/conll2003/deu.testb",
@@ -35,6 +42,8 @@ DOWNSTREAM_TASK_MAP = {
     'cola': "https://s3.eu-central-1.amazonaws.com/deepset.ai-farm-downstream/cola.tar.gz",
     "asnq_binary": "https://s3.eu-central-1.amazonaws.com/deepset.ai-farm-downstream/asnq_binary.tar.gz",
     "germeval17": "https://s3.eu-central-1.amazonaws.com/deepset.ai-farm-downstream/germeval17.tar.gz",
+    "natural_questions": "https://s3.eu-central-1.amazonaws.com/deepset.ai-farm-downstream/natural_questions.tar.gz",
+
 }
 
 def read_tsv(filename, rename_columns, quotechar='"', delimiter="\t", skiprows=None, header=0, proxies=None, max_samples=None):
@@ -45,7 +54,8 @@ def read_tsv(filename, rename_columns, quotechar='"', delimiter="\t", skiprows=N
         logger.info(f" Couldn't find {filename} locally. Trying to download ...")
         _download_extract_downstream_data(filename, proxies=proxies)
 
-    # read file into df
+    # read file into df - but only read those cols we need
+    columns_needed = list(rename_columns.keys())
     df = pd.read_csv(
         filename,
         sep=delimiter,
@@ -53,7 +63,8 @@ def read_tsv(filename, rename_columns, quotechar='"', delimiter="\t", skiprows=N
         quotechar=quotechar,
         dtype=str,
         skiprows=skiprows,
-        header=header
+        header=header,
+        usecols=columns_needed,
     )
     if max_samples:
         df = df.sample(max_samples)
@@ -61,11 +72,9 @@ def read_tsv(filename, rename_columns, quotechar='"', delimiter="\t", skiprows=N
     # let's rename our target columns to the default names FARM expects:
     # "text": contains the text
     # "text_classification_label": contains a label for text classification
-    columns = ["text"] + list(rename_columns.keys())
-    df = df[columns]
-    for source_column, label_name in rename_columns.items():
-        df[label_name] = df[source_column].fillna("")
-        df.drop(columns=[source_column], inplace=True)
+    df.rename(columns=rename_columns, inplace=True)
+    df.fillna("", inplace=True)
+
     # convert df to one dict per row
     raw_dict = df.to_dict(orient="records")
     return raw_dict
@@ -102,6 +111,14 @@ def read_tsv_sentence_pair(filename, rename_columns, delimiter="\t", skiprows=No
     raw_dict = df.to_dict(orient="records")
     return raw_dict
 
+def read_jsonl(file, proxies=None):
+    # get remote dataset if needed
+    if not (os.path.exists(file)):
+        logger.info(f" Couldn't find {file} locally. Trying to download ...")
+        _download_extract_downstream_data(file, proxies=proxies)
+    dicts = [json.loads(l) for l in open(file, encoding="utf-8")]
+    return dicts
+
 def read_ner_file(filename, sep="\t", proxies=None):
     """
     read file
@@ -166,6 +183,63 @@ def read_ner_file(filename, sep="\t", proxies=None):
         data.append({"text": " ".join(sentence), "ner_label": label})
     return data
 
+def read_dpr_json(file, max_samples=None, proxies=None):
+    """
+    Reads a Dense Passage Retrieval (DPR) data file in json format and returns a list of dictionaries.
+
+    :param file: filename of DPR data in json format
+
+    Returns:
+        list of dictionaries: List[dict]
+        each dictionary: {
+                    "query": str -> query_text
+                    "passages": List[dictionaries] -> [{"text": document_text, "title": xxx, "label": "positive", "external_id": abb123},
+                                {"text": document_text, "title": xxx, "label": "hard_negative", "external_id": abb134},
+                                ...]
+                    }
+        example:
+                ["query": 'who sings does he love me with reba'
+                "passages" : [{'title': 'Does He Love You',
+                    'text': 'Does He Love You "Does He Love You" is a song written by Sandy Knox and Billy Stritch, and recorded as a duet by American country music artists Reba McEntire and Linda Davis. It was released in August 1993 as the first single from Reba\'s album "Greatest Hits Volume Two". It is one of country music\'s several songs about a love triangle. "Does He Love You" was written in 1982 by Billy Stritch. He recorded it with a trio in which he performed at the time, because he wanted a song that could be sung by the other two members',
+                    'label': 'positive',
+                    'external_id': '11828866'},
+                    {'title': 'When the Nightingale Sings',
+                    'text': "When the Nightingale Sings When The Nightingale Sings is a Middle English poem, author unknown, recorded in the British Library's Harley 2253 manuscript, verse 25. It is a love poem, extolling the beauty and lost love of an unknown maiden. When þe nyhtegale singes þe wodes waxen grene.<br> Lef ant gras ant blosme springes in aueryl y wene,<br> Ant love is to myn herte gon wiþ one spere so kene<br> Nyht ant day my blod hit drynkes myn herte deþ me tene. Ich have loved al þis er þat y may love namore,<br> Ich have siked moni syk lemmon for",
+                    'label': 'hard_negative',
+                    'external_id': '10891637'}]
+                ]
+
+    """
+    # get remote dataset if needed
+    if not (os.path.exists(file)):
+        logger.info(f" Couldn't find {file} locally. Trying to download ...")
+        _download_extract_downstream_data(file, proxies=proxies)
+    dicts = json.load(open(file))
+    if max_samples:
+        dicts = random.sample(dicts, min(max_samples, len(dicts)))
+    # convert DPR dictionary to standard dictionary
+    query_json_keys = ["question", "questions", "query"]
+    positive_context_json_keys = ["positive_contexts", "positive_ctxs", "positive_context", "positive_ctx"]
+    hard_negative_json_keys = ["hard_negative_contexts", "hard_negative_ctxs", "hard_negative_context", "hard_negative_ctx"]
+    standard_dicts = []
+    for dict in dicts:
+        sample = {}
+        passages = []
+        for key, val in dict.items():
+            if key in query_json_keys:
+                sample["query"] = val
+            elif key in positive_context_json_keys+hard_negative_json_keys:
+                for passage in val:
+                    passages.append({
+                        "title": passage["title"],
+                        "text": passage["text"],
+                        "label": "positive" if key in positive_context_json_keys else "hard_negative",
+                        "external_id": passage["passage_id"]
+                        })
+        sample["passages"] = passages
+        standard_dicts.append(sample)
+    return standard_dicts
+
 def _convert_germeval14_labels(tags: List[str]):
     newtags = []
     for tag in tags:
@@ -212,11 +286,12 @@ def read_squad_file(filename, proxies=None):
 
 def write_squad_predictions(predictions, out_filename, predictions_filename=None):
     predictions_json = {}
-    for p in predictions:
-        if p["preds"][0][0] is not None:
-            predictions_json[p["id"]] = p["preds"][0][0]
-        else:
-            predictions_json[p["id"]] = "" #convert No answer = None to format understood by the SQuAD eval script
+    for x in predictions:
+        for p in x["predictions"]:
+            if p["answers"][0]["answer"] is not None:
+                predictions_json[p["question_id"]] = p["answers"][0]["answer"]
+            else:
+                predictions_json[p["question_id"]] = "" #convert No answer = None to format understood by the SQuAD eval script
 
     if predictions_filename:
         dev_labels = {}
@@ -224,20 +299,20 @@ def write_squad_predictions(predictions, out_filename, predictions_filename=None
         for d in temp["data"]:
             for p in d["paragraphs"]:
                 for q in p["qas"]:
-                    if q["is_impossible"]:
+                    if q.get("is_impossible",False):
                         dev_labels[q["id"]] = "is_impossible"
                     else:
                         dev_labels[q["id"]] = q["answers"][0]["text"]
         not_included = set(list(dev_labels.keys())) - set(list(predictions_json.keys()))
         if len(not_included) > 0:
-            logger.info(f"There were missing predicitons for question ids: {str(set(list(dev_labels.keys())))}")
+            logger.info(f"There were missing predicitons for question ids: {list(not_included)}")
         for x in not_included:
             predictions_json[x] = ""
 
-    os.makedirs("model_output", exist_ok=True)
-    filepath = Path("model_output") / out_filename
-    json.dump(predictions_json, open(filepath, "w"))
-    logger.info(f"Written Squad predictions to: {filepath}")
+    # os.makedirs("model_output", exist_ok=True)
+    # filepath = Path("model_output") / out_filename
+    json.dump(predictions_json, open(out_filename, "w"))
+    logger.info(f"Written Squad predictions to: {out_filename}")
 
 def _get_md5checksum(fname):
     # solution from stackoverflow: https://stackoverflow.com/a/3431838
@@ -324,7 +399,7 @@ def _conll03get(dataset, directory, language):
 
 
 
-def read_docs_from_txt(filename, delimiter="", encoding="utf-8", max_docs=None, proxies=None):
+def read_docs_from_txt(filename, delimiter="", encoding="utf-8", max_docs=None, proxies=None, disable_tqdm=True):
     """Reads a text file with one sentence per line and a delimiter between docs (default: empty lines) ."""
     if not (os.path.exists(filename)):
         _download_extract_downstream_data(filename, proxies)
@@ -335,7 +410,7 @@ def read_docs_from_txt(filename, delimiter="", encoding="utf-8", max_docs=None,
     corpus_lines = 0
 
     with open(filename, "r", encoding=encoding) as f:
-        for line_num, line in enumerate(tqdm(f, desc="Loading Dataset", total=corpus_lines)):
+        for line_num, line in enumerate(tqdm(f, desc="Loading Dataset", total=corpus_lines, disable=disable_tqdm)):
             line = line.strip()
             if line == delimiter:
                 if len(doc) > 0:
@@ -348,7 +423,7 @@ def read_docs_from_txt(filename, delimiter="", encoding="utf-8", max_docs=None,
                             logger.info(f"Reached number of max_docs ({max_docs}). Skipping rest of file ...")
                             break
                 else:
-                    logger.warning(f"Found empty document in file (line {line_num}). "
+                    logger.warning(f"Found empty document in '{filename}' (line {line_num}). "
                                    f"Make sure that you comply with the format: "
                                    f"One sentence per line and exactly *one* empty line between docs. "
                                    f"You might have multiple subsequent empty lines.")
@@ -438,7 +513,7 @@ def _get_random_sentence(all_baskets, forbidden_doc):
     sentence = None
     for _ in range(100):
         rand_doc_idx = random.randrange(len(all_baskets))
-        rand_doc = all_baskets[rand_doc_idx]["doc"]
+        rand_doc = all_baskets[rand_doc_idx]
 
         # check if our picked random doc is really different to our initial doc
         if rand_doc != forbidden_doc:
@@ -450,102 +525,8 @@ def _get_random_sentence(all_baskets, forbidden_doc):
     return sentence
 
 
-def get_sequence_pair(doc, chunk, chunk_clear_text, all_baskets, tokenizer, max_num_tokens, prob_next_sentence=0.5):
-    """
-    Get one sample from corpus consisting of two sequences. A sequence can consist of more than one sentence.
-    With prob. 50% these are two subsequent sequences from one doc. With 50% the second sequence will be a
-    random one from another document.
 
-    :param doc: The current document.
-    :type doc: [str]
-    :param chunk: List of subsequent, tokenized sentences.
-    :type chunk: [dict]
-    :param chunk_clear_text: List of subsequent sentences.
-    :type chunk_clear_text: [str]
-    :param all_baskets: SampleBaskets containing multiple other docs from which we can sample the second sequence
-    if we need a random one.
-    :type all_baskets: [dict]
-    :param tokenizer: Used to split a sentence (str) into tokens.
-    :param max_num_tokens: Samples are truncated after this many tokens.
-    :type max_num_tokens: int
-    :return: (list, list, dict, int)
-        tokenized seq a,
-        tokenized seq b,
-        sample in clear text with label,
-        number of unused sentences in chunk
-    """
-    sequence_a = []
-    sequence_b = []
-    sample_in_clear_text = { "text_a" : "", "text_b" : ""}
-    # determine how many segments from chunk go into sequence_a
-    len_sequence_a = 0
-    a_end = 1
-    if len(chunk) >= 2:
-        a_end = random.randrange(1, len(chunk))
-    for i in range(a_end):
-        sequence_a.append(chunk[i])
-        sample_in_clear_text["text_a"] += f"{chunk_clear_text[i]} "
-        len_sequence_a += len(chunk[i]["tokens"])
-    sample_in_clear_text["text_a"].strip()
-
-    # actual next sequence
-    if (random.random() > prob_next_sentence) and (len(chunk) > 1):
-        label = True
-        for i in range(a_end, len(chunk)):
-            sequence_b.append(chunk[i])
-            sample_in_clear_text["text_b"] += f"{chunk_clear_text[i]} "
-        sample_in_clear_text["text_b"].strip()
-        sample_in_clear_text["nextsentence_label"] = True
-        num_unused_segments = 0
-    # edge case: split sequence in half
-    elif (len(chunk) == 1) and len_sequence_a >= max_num_tokens:
-        sequence_a = {}
-        sequence_b = {}
-        if int(len(chunk[0]["tokens"])/2) >= max_num_tokens:
-            boundary = int(max_num_tokens/2)
-        else:
-            boundary = int(len(chunk[0]["tokens"])/2)
-        sequence_a["tokens"] = chunk[0]["tokens"][:boundary]
-        sequence_a["offsets"] = chunk[0]["offsets"][:boundary]
-        sequence_a["start_of_word"] = chunk[0]["start_of_word"][:boundary]
-        sequence_b["tokens"] = chunk[0]["tokens"][boundary:]
-        sequence_b["start_of_word"] = chunk[0]["start_of_word"][boundary:]
-        # get offsets for sequence_b right
-        seq_b_offset_start = chunk[0]["offsets"][boundary]
-        sequence_b["offsets"] = [offset - seq_b_offset_start for offset in chunk[0]["offsets"][boundary:]]
-        # get clear text
-        clear_text_boundary = chunk[0]["offsets"][boundary]
-        sample_in_clear_text["text_a"] = chunk_clear_text[0][:clear_text_boundary]
-        sample_in_clear_text["text_b"] = chunk_clear_text[0][clear_text_boundary:]
-        sample_in_clear_text["text_a"].strip()
-        sample_in_clear_text["text_b"].strip()
-        sample_in_clear_text["nextsentence_label"] = True
-        return [sequence_a], [sequence_b], sample_in_clear_text, 0
-    # random next sequence
-    else:
-        label = False
-        sequence_b_length = 0
-        target_b_length = max_num_tokens - len_sequence_a
-        random_doc = _get_random_doc(all_baskets, forbidden_doc=doc)
-
-        random_start = random.randrange(len(random_doc))
-        for i in range(random_start, len(random_doc)):
-            current_sentence_tokenized = tokenize_with_metadata(random_doc[i], tokenizer)
-            sequence_b.append(current_sentence_tokenized)
-            sample_in_clear_text["text_b"] += f"{random_doc[i]} "
-            sequence_b_length += len(current_sentence_tokenized["tokens"])
-            if sequence_b_length >= target_b_length:
-                break
-
-        sample_in_clear_text["text_b"].strip()
-        sample_in_clear_text["nextsentence_label"] = False
-
-        # We didn't use all of the segments in chunk => put them back
-        num_unused_segments = len(chunk) - a_end
-
-    assert len(sequence_a) > 0
-    assert len(sequence_b) > 0
-    return sequence_a, sequence_b, sample_in_clear_text, num_unused_segments
+    # return sequence_a, sequence_b, sample_in_clear_text, num_unused_segments
 
 
 def _get_random_doc(all_baskets, forbidden_doc):
@@ -587,88 +568,7 @@ def join_sentences(sequence):
     return sequence_joined
 
 
-def mask_random_words(tokens, vocab, token_groups=None, max_predictions_per_seq=20, masked_lm_prob=0.15):
-    """
-    Masking some random tokens for Language Model task with probabilities as in the original BERT paper.
-    num_masked.
-    If token_groups is supplied, whole word masking is applied, so *all* tokens of a word are either masked or not.
-    This option was added by the BERT authors later and showed solid improvements compared to the original objective.
-    Whole Word Masking means that if we mask all of the wordpieces corresponding to an original word.
-    When a word has been split intoWordPieces, the first token does not have any marker and any subsequence
-    tokens are prefixed with ##. So whenever we see the ## token, we
-    append it to the previous set of word indexes. Note that Whole Word Masking does *not* change the training code
-    at all -- we still predict each WordPiece independently, softmaxed over the entire vocabulary.
-    This implementation is mainly a copy from the original code by Google, but includes some simplifications.
 
-    :param tokens: tokenized sentence.
-    :type tokens: [str]
-    :param vocab: vocabulary for choosing tokens for random masking.
-    :type vocab: dict
-    :param token_groups: If supplied, only whole groups of tokens get masked. This can be whole words but
-    also other types (e.g. spans). Booleans indicate the start of a group.
-    :type token_groups: [bool]
-    :param max_predictions_per_seq: maximum number of masked tokens
-    :type max_predictions_per_seq: int
-    :param masked_lm_prob: probability of masking a token
-    :type masked_lm_prob: float
-    :return: (list of str, list of int), masked tokens and related labels for LM prediction
-    """
-
-    #TODO make special tokens model independent
-
-    # 1. Combine tokens to one group (e.g. all subtokens of a word)
-    cand_indices = []
-    for (i, token) in enumerate(tokens):
-        if token == "[CLS]" or token == "[SEP]":
-            continue
-        if (token_groups and len(cand_indices) >= 1 and not token_groups[i]):
-            cand_indices[-1].append(i)
-        else:
-            cand_indices.append([i])
-
-    num_to_mask = min(max_predictions_per_seq,
-                      max(1, int(round(len(tokens) * masked_lm_prob))))
-
-    random.shuffle(cand_indices)
-    output_label = [''] * len(tokens)
-    num_masked = 0
-    assert "[MASK]" not in tokens
-
-    # 2. Mask the first groups until we reach the number of tokens we wanted to mask (num_to_mask)
-    for index_set in cand_indices:
-        if num_masked >= num_to_mask:
-            break
-        # If adding a whole-word mask would exceed the maximum number of
-        # predictions, then just skip this candidate.
-        if num_masked + len(index_set) > num_to_mask:
-            continue
-
-        for index in index_set:
-            prob = random.random()
-            num_masked += 1
-            original_token = tokens[index]
-            # 80% randomly change token to mask token
-            if prob < 0.8:
-                tokens[index] = "[MASK]"
-
-            # 10% randomly change token to random token
-            #TODO currently custom vocab is not included here
-            elif prob < 0.9:
-                tokens[index] = random.choice(list(vocab.items()))[0]
-
-            # -> rest 10% randomly keep current token
-
-            # append current token to output (we will predict these later)
-            try:
-                output_label[index] = original_token
-            except KeyError:
-                # For unknown words (should not occur with BPE vocab)
-                output_label[index] = "[UNK]"
-                logger.warning(
-                    "Cannot find token '{}' in vocab. Using [UNK] instead".format(original_token)
-                )
-
-    return tokens, output_label
 
 
 def is_json(x):
@@ -706,7 +606,7 @@ def grouper(iterable, n, worker_id=0, total_workers=1):
     Output for worker 2: [(dictC, dictD), (dictI, dictJ), ...]
     Output for worker 3: [(dictE, dictF), (dictK, dictL), ...]
 
-    This method also adds an index number to every dict yielded similar to the grouper().
+    This method also adds an index number to every dict yielded.
 
     :param iterable: a generator object that yields dicts
     :type iterable: generator
@@ -747,3 +647,58 @@ def grouper(iterable, n, worker_id=0, total_workers=1):
         iterable = filter_elements_per_worker(iterable)
 
     return iter(lambda: list(islice(iterable, n)), [])
+
+
+def split_file(filepath, output_dir, docs_per_file=1_000, delimiter="", encoding="utf-8"):
+    total_lines = sum(1 for line in open(filepath, encoding=encoding))
+    output_file_number = 1
+    doc_count = 0
+    lines_to_write = []
+    with ExitStack() as stack:
+        input_file = stack.enter_context(open(filepath, 'r', encoding=encoding))
+        for line_num, line in enumerate(tqdm(input_file, desc="Splitting file ...", total=total_lines)):
+            lines_to_write.append(line)
+            if line.strip() == delimiter:
+                doc_count += 1
+                if doc_count % docs_per_file == 0:
+                    filename = output_dir / f"part_{output_file_number}"
+                    os.makedirs(os.path.dirname(filename), exist_ok=True)
+                    write_file = stack.enter_context(open(filename, 'w+', encoding=encoding, buffering=10 * 1024 * 1024))
+                    write_file.writelines(lines_to_write)
+                    write_file.close()
+                    output_file_number += 1
+                    lines_to_write = []
+
+        if lines_to_write:
+            filename = output_dir / f"part_{output_file_number}"
+            os.makedirs(os.path.dirname(filename), exist_ok=True)
+            write_file = stack.enter_context(open(filename, 'w+', encoding=encoding, buffering=10 * 1024 * 1024))
+            write_file.writelines(lines_to_write)
+            write_file.close()
+
+    logger.info(f"The input file {filepath} is split in {output_file_number} parts at {output_dir}.")
+
+
+def generate_tok_to_ch_map(text):
+    """ Generates a mapping from token to character index when a string text is split using .split()
+    TODO e.g."""
+    map = [0]
+    follows_whitespace = False
+    for i, ch in enumerate(text):
+        if follows_whitespace:
+            if ch not in string.whitespace:
+                map.append(i)
+                follows_whitespace = False
+        else:
+            if ch in string.whitespace:
+                follows_whitespace = True
+    return map
+
+
+def split_with_metadata(text):
+    """" Splits a string text by whitespace and also returns indexes which is a mapping from token index
+    to character index"""
+    split_text = text.split()
+    indexes = generate_tok_to_ch_map(text)
+    assert len(split_text) == len(indexes)
+    return split_text, indexes
