@@ -2,24 +2,25 @@ import logging
 import multiprocessing as mp
 import os
 from functools import partial
+import warnings
 
 import torch
 from torch.utils.data.sampler import SequentialSampler
 from tqdm import tqdm
-from transformers.configuration_auto import AutoConfig
+from typing import Generator, List, Union
 
 from farm.data_handler.dataloader import NamedDataLoader
-from farm.data_handler.processor import Processor, InferenceProcessor, SquadProcessor, NERProcessor, TextClassificationProcessor
+from farm.data_handler.processor import Processor, InferenceProcessor, NaturalQuestionsProcessor
 from farm.data_handler.utils import grouper
-from farm.modeling.tokenization import Tokenizer
-from farm.modeling.adaptive_model import AdaptiveModel, BaseAdaptiveModel
-from farm.utils import initialize_device_settings
-from farm.utils import set_all_seeds, calc_chunksize, log_ascii_workers
-
+from farm.data_handler.inputs import QAInput
+from farm.modeling.adaptive_model import AdaptiveModel, BaseAdaptiveModel, ONNXAdaptiveModel
+from farm.modeling.optimization import optimize_model
+from farm.utils import initialize_device_settings, MLFlowLogger
+from farm.utils import set_all_seeds, calc_chunksize, log_ascii_workers, Benchmarker
+from farm.modeling.predictions import QAPred
 
 logger = logging.getLogger(__name__)
 
-
 class Inferencer:
     """
     Loads a saved AdaptiveModel/ONNXAdaptiveModel from disk and runs it in inference mode. Can be used for a
@@ -54,7 +55,9 @@ class Inferencer:
         extraction_layer=None,
         s3e_stats=None,
         num_processes=None,
-        disable_tqdm=False
+        disable_tqdm=False,
+        benchmarking=False,
+        dummy_ph=False
     ):
         """
         Initializes Inferencer from an AdaptiveModel and a Processor instance.
@@ -83,15 +86,38 @@ class Inferencer:
         :param s3e_stats: Stats of a fitted S3E model as returned by `fit_s3e_on_corpus()`
                           (only needed for task_type="embeddings" and extraction_strategy = "s3e")
         :type s3e_stats: dict
-        :param num_processes: the number of processes for `multiprocessing.Pool`. Set to value of 0 to disable
-                              multiprocessing. Set to None to let Inferencer use all CPU cores. If you want to
-                              debug the Language Model, you might need to disable multiprocessing!
+        :param num_processes: the number of processes for `multiprocessing.Pool`.
+                              Set to value of 1 (or 0) to disable multiprocessing.
+                              Set to None to let Inferencer use all CPU cores minus one.
+                              If you want to debug the Language Model, you might need to disable multiprocessing!
+                              **Warning!** If you use multiprocessing you have to close the
+                              `multiprocessing.Pool` again! To do so call
+                              :func:`~farm.infer.Inferencer.close_multiprocessing_pool` after you are
+                              done using this class. The garbage collector will not do this for you!
         :type num_processes: int
         :param disable_tqdm: Whether to disable tqdm logging (can get very verbose in multiprocessing)
         :type disable_tqdm: bool
+        :param dummy_ph: If True, methods of the prediction head will be replaced
+                     with a dummy method. This is used to isolate lm run time from ph run time.
+        :type dummy_ph: bool
+        :param benchmarking: If True, a benchmarking object will be initialised within the class and
+                             certain parts of the code will be timed for benchmarking. Should be kept
+                             False if not benchmarking since these timing checkpoints require synchronization
+                             of the asynchronous Pytorch operations and may slow down the model.
+        :type benchmarking: bool
         :return: An instance of the Inferencer.
 
         """
+        MLFlowLogger.disable()
+
+        # For benchmarking
+        if dummy_ph:
+            model.bypass_ph()
+
+        self.benchmarking = benchmarking
+        if self.benchmarking:
+            self.benchmarker = Benchmarker()
+
         # Init device and distributed settings
         device, n_gpu = initialize_device_settings(use_cuda=gpu, local_rank=-1, use_amp=None)
 
@@ -103,6 +129,7 @@ class Inferencer:
         self.language = self.model.get_language()
         self.task_type = task_type
         self.disable_tqdm = disable_tqdm
+        self.problematic_sample_ids = set()
 
         if task_type == "embeddings":
             if not extraction_layer or not extraction_strategy:
@@ -127,6 +154,7 @@ class Inferencer:
     def load(
         cls,
         model_name_or_path,
+        revision=None,
         batch_size=4,
         gpu=False,
         task_type=None,
@@ -138,8 +166,13 @@ class Inferencer:
         extraction_strategy=None,
         s3e_stats=None,
         num_processes=None,
-        disable_tqdm=False
-
+        disable_tqdm=False,
+        tokenizer_class=None,
+        use_fast=True,
+        tokenizer_args=None,
+        multithreading_rust=True,
+        dummy_ph=False,
+        benchmarking=False,
     ):
         """
         Load an Inferencer incl. all relevant components (model, tokenizer, processor ...) either by
@@ -149,6 +182,8 @@ class Inferencer:
 
         :param model_name_or_path: Local directory or public name of the model to load.
         :type model_name_or_path: str
+        :param revision: The version of model to use from the HuggingFace model hub. Can be tag name, branch name, or commit hash.
+        :type revision: str
         :param batch_size: Number of samples computed once per batch
         :type batch_size: int
         :param gpu: If GPU shall be used
@@ -173,14 +208,41 @@ class Inferencer:
                           (only needed for task_type="embeddings" and extraction_strategy = "s3e")
         :type s3e_stats: dict
         :param num_processes: the number of processes for `multiprocessing.Pool`. Set to value of 0 to disable
-                              multiprocessing. Set to None to let Inferencer use all CPU cores. If you want to
+                              multiprocessing. Set to None to let Inferencer use all CPU cores minus one. If you want to
                               debug the Language Model, you might need to disable multiprocessing!
+                              **Warning!** If you use multiprocessing you have to close the
+                              `multiprocessing.Pool` again! To do so call
+                              :func:`~farm.infer.Inferencer.close_multiprocessing_pool` after you are
+                              done using this class. The garbage collector will not do this for you!
         :type num_processes: int
         :param disable_tqdm: Whether to disable tqdm logging (can get very verbose in multiprocessing)
         :type disable_tqdm: bool
+        :param tokenizer_class: (Optional) Name of the tokenizer class to load (e.g. `BertTokenizer`)
+        :type tokenizer_class: str
+        :param use_fast: (Optional, True by default) Indicate if FARM should try to load the fast version of the tokenizer (True) or
+            use the Python one (False).
+        :type use_fast: bool
+        :param tokenizer_args: (Optional) Will be passed to the Tokenizer ``__init__`` method.
+            See https://huggingface.co/transformers/main_classes/tokenizer.html and detailed tokenizer documentation
+            on `Hugging Face Transformers <https://huggingface.co/transformers/>`_.
+        :type tokenizer_args: dict
+        :param multithreading_rust: Whether to allow multithreading in Rust, e.g. for FastTokenizers.
+                                    Note: Enabling multithreading in Rust AND multiprocessing in python might cause
+                                    deadlocks.
+        :type multithreading_rust: bool
+        :param dummy_ph: If True, methods of the prediction head will be replaced
+                             with a dummy method. This is used to isolate lm run time from ph run time.
+        :type dummy_ph: bool
+        :param benchmarking: If True, a benchmarking object will be initialised within the class and
+                             certain parts of the code will be timed for benchmarking. Should be kept
+                             False if not benchmarking since these timing checkpoints require synchronization
+                             of the asynchronous Pytorch operations and may slow down the model.
+        :type benchmarking: bool
         :return: An instance of the Inferencer.
 
         """
+        if tokenizer_args is None:
+            tokenizer_args = {}
 
         device, n_gpu = initialize_device_settings(use_cuda=gpu, local_rank=-1, use_amp=None)
         name = os.path.basename(model_name_or_path)
@@ -193,56 +255,34 @@ class Inferencer:
             else:
                 processor = Processor.load_from_dir(model_name_or_path)
 
-            # override processor attributes loaded from config file with inferencer params
-            processor.max_seq_len = max_seq_len
-            if hasattr(processor, "doc_stride"):
-                processor.doc_stride = doc_stride
-
         # b) or from remote transformers model hub
         else:
-            logger.info(f"Could not find `{model_name_or_path}` locally. Try to download from model hub ...")
             if not task_type:
                 raise ValueError("Please specify the 'task_type' of the model you want to load from transformers. "
                                  "Valid options for arg `task_type`:"
                                  "'question_answering', 'embeddings', 'text_classification', 'ner'")
 
-            model = AdaptiveModel.convert_from_transformers(model_name_or_path, device, task_type)
-            config = AutoConfig.from_pretrained(model_name_or_path)
-            tokenizer = Tokenizer.load(model_name_or_path)
+            model = AdaptiveModel.convert_from_transformers(model_name_or_path,
+                                                            revision=revision,
+                                                            device=device,
+                                                            task_type=task_type)
+            processor = Processor.convert_from_transformers(model_name_or_path,
+                                                            revision=revision,
+                                                            task_type=task_type,
+                                                            max_seq_len=max_seq_len,
+                                                            doc_stride=doc_stride,
+                                                            tokenizer_class=tokenizer_class,
+                                                            tokenizer_args=tokenizer_args,
+                                                            use_fast=use_fast)
 
-            # TODO infer task_type automatically from config (if possible)
-            if task_type == "question_answering":
-                processor = SquadProcessor(
-                    tokenizer=tokenizer,
-                    max_seq_len=max_seq_len,
-                    label_list=["start_token", "end_token"],
-                    metric="squad",
-                    data_dir="data",
-                    doc_stride=doc_stride
-                )
-            elif task_type == "embeddings":
-                processor = InferenceProcessor(tokenizer=tokenizer, max_seq_len=max_seq_len)
-
-            elif task_type == "text_classification":
-                label_list = list(config.id2label[id] for id in range(len(config.id2label)))
-                processor = TextClassificationProcessor(tokenizer=tokenizer,
-                                                        max_seq_len=max_seq_len,
-                                                        data_dir="data",
-                                                        label_list=label_list,
-                                                        label_column_name="label",
-                                                        metric="acc",
-                                                        quote_char='"',
-                                                        )
-            elif task_type == "ner":
-                label_list = list(config.label2id.keys())
-                processor = NERProcessor(
-                    tokenizer=tokenizer, max_seq_len=max_seq_len, data_dir="data", metric="seq_f1",
-                    label_list=label_list
-                )
-            else:
-                raise ValueError(f"`task_type` {task_type} is not supported yet. "
-                                 f"Valid options for arg `task_type`: 'question_answering', "
-                                 f"'embeddings', 'text_classification', 'ner'")
+        # override processor attributes loaded from config or HF with inferencer params
+        processor.max_seq_len = max_seq_len
+        processor.multithreading_rust = multithreading_rust
+        if hasattr(processor, "doc_stride"):
+            assert doc_stride < max_seq_len, "doc_stride is longer than max_seq_len. This means that there will be gaps " \
+                                             "as the passage windows slide, causing the model to skip over parts of the document. " \
+                                             "Please set a lower value for doc_stride (Suggestions: doc_stride=128, max_seq_len=384) "
+            processor.doc_stride = doc_stride
 
         return cls(
             model,
@@ -256,36 +296,62 @@ class Inferencer:
             extraction_layer=extraction_layer,
             s3e_stats=s3e_stats,
             num_processes=num_processes,
-            disable_tqdm=disable_tqdm
+            disable_tqdm=disable_tqdm,
+            benchmarking=benchmarking,
+            dummy_ph=dummy_ph
         )
 
     def _set_multiprocessing_pool(self, num_processes):
         """
         Initialize a multiprocessing.Pool for instances of Inferencer.
 
-         :param num_processes: the number of processes for `multiprocessing.Pool`. Set to value of 0 to disable
-                               multiprocessing. Set to None to let Inferencer use all CPU cores. If you want to
-                               debug the Language Model, you might need to disable multiprocessing!
+        :param num_processes: the number of processes for `multiprocessing.Pool`.
+                              Set to value of 1 (or 0) to disable multiprocessing.
+                              Set to None to let Inferencer use all CPU cores minus one.
+                              If you want to debug the Language Model, you might need to disable multiprocessing!
+                              **Warning!** If you use multiprocessing you have to close the
+                              `multiprocessing.Pool` again! To do so call
+                              :func:`~farm.infer.Inferencer.close_multiprocessing_pool` after you are
+                              done using this class. The garbage collector will not do this for you!
         :type num_processes: int
         :return:
         """
         self.process_pool = None
-        if num_processes == 0:  # disable multiprocessing
+        if num_processes == 0 or num_processes == 1:  # disable multiprocessing
             self.process_pool = None
         else:
             if num_processes is None:  # use all CPU cores
-                num_processes = mp.cpu_count() - 1
+                if mp.cpu_count() > 3:
+                    num_processes = mp.cpu_count() - 1
+                else:
+                    num_processes = mp.cpu_count()
             self.process_pool = mp.Pool(processes=num_processes)
             logger.info(
                 f"Got ya {num_processes} parallel workers to do inference ..."
             )
             log_ascii_workers(n=num_processes,logger=logger)
 
+    def close_multiprocessing_pool(self, join=False):
+        """Close the `multiprocessing.Pool` again.
+
+        If you use multiprocessing you have to close the `multiprocessing.Pool` again!
+        To do so call this function after you are done using this class.
+        The garbage collector will not do this for you!
+
+        :param join: wait for the worker processes to exit
+        :type join: bool
+        """
+        if self.process_pool is not None:
+            self.process_pool.close()
+            if join:
+                self.process_pool.join()
+            self.process_pool = None
+
     def save(self, path):
         self.model.save(path)
         self.processor.save(path)
 
-    def inference_from_file(self, file, multiprocessing_chunksize=None, streaming=False):
+    def inference_from_file(self, file, multiprocessing_chunksize=None, streaming=False, return_json=True):
         """
         Run down-stream inference on samples created from an input file.
         The file should be in the same format as the ones used during training
@@ -307,7 +373,7 @@ class Inferencer:
         dicts = self.processor.file_to_dicts(file)
         preds_all = self.inference_from_dicts(
             dicts,
-            rest_api_schema=False,
+            return_json=return_json,
             multiprocessing_chunksize=multiprocessing_chunksize,
             streaming=streaming,
         )
@@ -317,16 +383,15 @@ class Inferencer:
             return list(preds_all)
 
     def inference_from_dicts(
-        self, dicts, rest_api_schema=False, multiprocessing_chunksize=None, streaming=False
+        self, dicts, return_json=True, multiprocessing_chunksize=None, streaming=False
     ):
         """
         Runs down-stream inference on samples created from input dictionaries.
         The format of the input `dicts` depends on the task:
 
-        * QA: [{"qas": ["What is X?"], "context":  "Some context containing the answer"}]
+        * QA (FARM style): [{"questions": ["What is X?"], "text":  "Some context containing the answer"}]
         * Classification / NER / embeddings: [{"text": "Some input text"}]
 
-
         Inferencer has a high performance non-blocking streaming mode for large scale inference use cases. With this
         mode, the dicts parameter can optionally be a Python generator object that yield dicts, thus avoiding loading
         dicts in memory. The inference_from_dicts() method returns a generator that yield predictions. To use streaming,
@@ -336,10 +401,9 @@ class Inferencer:
         :param dicts: Samples to run inference on provided as a list(or a generator object) of dicts.
                       One dict per sample.
         :type dicts: iter(dict)
-        :param rest_api_schema: Whether input dicts use the format that complies with the FARM REST API.
-                                Currently only used for QA to switch from squad to a more useful format in production.
-                                While input is almost the same, output contains additional meta data(offset, context..)
-        :type rest_api_schema: bool
+        :param return_json: Whether the output should be in a json appropriate format. If False, it returns the prediction
+                            object where applicable, else it returns PredObj.to_json()
+        :type return_json: bool
         :return: dict of predictions
         :param multiprocessing_chunksize: number of dicts to put together in one chunk and feed to one process
                                           (only relevant if you do multiprocessing)
@@ -355,12 +419,17 @@ class Inferencer:
         """
 
         # whether to aggregate predictions across different samples (e.g. for QA on long texts)
+        # TODO remove or adjust after implmenting input objects properly
+        # if set(dicts[0].keys()) == {"qas", "context"}:
+        #     warnings.warn("QA Input dictionaries with [qas, context] as keys will be deprecated in the future",
+        #                   DeprecationWarning)
+
         aggregate_preds = False
         if len(self.model.prediction_heads) > 0:
             aggregate_preds = hasattr(self.model.prediction_heads[0], "aggregate_preds")
 
         if self.process_pool is None:  # multiprocessing disabled (helpful for debugging or using in web frameworks)
-            predictions = self._inference_without_multiprocessing(dicts, rest_api_schema, aggregate_preds)
+            predictions = self._inference_without_multiprocessing(dicts, return_json, aggregate_preds)
             return predictions
         else:  # use multiprocessing for inference
             # Calculate values of multiprocessing_chunksize and num_processes if not supplied in the parameters.
@@ -378,54 +447,65 @@ class Inferencer:
                     multiprocessing_chunksize = _chunk_size
 
             predictions = self._inference_with_multiprocessing(
-                dicts, rest_api_schema, aggregate_preds, multiprocessing_chunksize,
+                dicts, return_json, aggregate_preds, multiprocessing_chunksize,
             )
 
+            self.processor.log_problematic(self.problematic_sample_ids)
             # return a generator object if streaming is enabled, else, cast the generator to a list.
             if not streaming and type(predictions) != list:
                 return list(predictions)
             else:
                 return predictions
 
-    def _inference_without_multiprocessing(self, dicts, rest_api_schema, aggregate_preds):
+    def _inference_without_multiprocessing(self, dicts, return_json, aggregate_preds):
         """
         Implementation of inference from dicts without using Python multiprocessing. Useful for debugging or in API
         framework where spawning new processes could be expensive.
 
         :param dicts: Samples to run inference on provided as a list of dicts. One dict per sample.
         :type dicts: iter(dict)
-        :param rest_api_schema: Whether input dicts use the format that complies with the FARM REST API.
-                                Currently only used for QA to switch from squad to a more useful format in production.
-                                While input is almost the same, output contains additional meta data(offset, context..)
-        :type rest_api_schema: bool
+        :param return_json: Whether the output should be in a json appropriate format. If False, it returns the prediction
+                            object where applicable, else it returns PredObj.to_json()
+        :type return_json: bool
         :param aggregate_preds: whether to aggregate predictions across different samples (e.g. for QA on long texts)
         :type aggregate_preds: bool
 
         :return: list of predictions
         :rtype: list
         """
-        dataset, tensor_names, baskets = self.processor.dataset_from_dicts(
-            dicts, indices=[i for i in range(len(dicts))], rest_api_schema=rest_api_schema, return_baskets=True
+        dataset, tensor_names, problematic_ids, baskets = self.processor.dataset_from_dicts(
+            dicts, indices=[i for i in range(len(dicts))], return_baskets=True
         )
+        self.problematic_sample_ids = problematic_ids
+        if self.benchmarking:
+            self.benchmarker.record("dataset_single_proc")
+
         # TODO change format of formatted_preds in QA (list of dicts)
         if aggregate_preds:
-            preds_all = self._get_predictions_and_aggregate(dataset, tensor_names, baskets, rest_api_schema)
+            preds_all = self._get_predictions_and_aggregate(dataset, tensor_names, baskets)
         else:
-            preds_all = self._get_predictions(dataset, tensor_names, baskets, rest_api_schema)
+            preds_all = self._get_predictions(dataset, tensor_names, baskets)
+
+        if return_json:
+            # TODO this try catch should be removed when all tasks return prediction objects
+            try:
+                preds_all = [x.to_json() for x in preds_all]
+            except AttributeError:
+                pass
+
         return preds_all
 
     def _inference_with_multiprocessing(
-        self, dicts, rest_api_schema, aggregate_preds, multiprocessing_chunksize
+        self, dicts, return_json, aggregate_preds, multiprocessing_chunksize
     ):
         """
         Implementation of inference. This method is a generator that yields the results.
 
         :param dicts: Samples to run inference on provided as a list of dicts or a generator object that yield dicts.
         :type dicts: iter(dict)
-        :param rest_api_schema: Whether input dicts use the format that complies with the FARM REST API.
-                                Currently only used for QA to switch from squad to a more useful format in production.
-                                While input is almost the same, output contains additional meta data(offset, context..)
-        :type rest_api_schema: bool
+        :param return_json: Whether the output should be in a json appropriate format. If False, it returns the prediction
+                            object where applicable, else it returns PredObj.to_json()
+        :type return_json: bool
         :param aggregate_preds: whether to aggregate predictions across different samples (e.g. for QA on long texts)
         :type aggregate_preds: bool
         :param multiprocessing_chunksize: number of dicts to put together in one chunk and feed to one process
@@ -437,34 +517,46 @@ class Inferencer:
         # We group the input dicts into chunks and feed each chunk to a different process
         # in the pool, where it gets converted to a pytorch dataset
         results = self.process_pool.imap(
-            partial(self._create_datasets_chunkwise, processor=self.processor, rest_api_schema=rest_api_schema),
+            partial(self._create_datasets_chunkwise, processor=self.processor),
             grouper(iterable=dicts, n=multiprocessing_chunksize),
             1,
         )
 
         # Once a process spits out a preprocessed chunk. we feed this dataset directly to the model.
         # So we don't need to wait until all preprocessing has finished before getting first predictions.
-        for dataset, tensor_names, baskets in results:
-            # TODO change format of formatted_preds in QA (list of dicts)
-            if aggregate_preds:
-                predictions = self._get_predictions_and_aggregate(
-                    dataset, tensor_names, baskets, rest_api_schema
-                )
+        for dataset, tensor_names, problematic_sample_ids, baskets in results:
+            self.problematic_sample_ids.update(problematic_sample_ids)
+            if dataset is None:
+                logger.error(f"Part of the dataset could not be converted! \n"
+                             f"BE AWARE: The order of predictions will not conform with the input order!")
             else:
-                predictions = self._get_predictions(dataset, tensor_names, baskets, rest_api_schema)
-            yield from predictions
+                # TODO change format of formatted_preds in QA (list of dicts)
+                if aggregate_preds:
+                    predictions = self._get_predictions_and_aggregate(
+                        dataset, tensor_names, baskets
+                    )
+                else:
+                    predictions = self._get_predictions(dataset, tensor_names, baskets)
+
+                if return_json:
+                    # TODO this try catch should be removed when all tasks return prediction objects
+                    try:
+                        predictions = [x.to_json() for x in predictions]
+                    except AttributeError:
+                        pass
+                yield from predictions
 
     @classmethod
-    def _create_datasets_chunkwise(cls, chunk, processor, rest_api_schema):
+    def _create_datasets_chunkwise(cls, chunk, processor):
         """Convert ONE chunk of data (i.e. dictionaries) into ONE pytorch dataset.
         This is usually executed in one of many parallel processes.
         The resulting datasets of the processes are merged together afterwards"""
         dicts = [d[1] for d in chunk]
         indices = [d[0] for d in chunk]
-        dataset, tensor_names, baskets = processor.dataset_from_dicts(dicts, indices, rest_api_schema, return_baskets=True)
-        return dataset, tensor_names, baskets
+        dataset, tensor_names, problematic_sample_ids, baskets = processor.dataset_from_dicts(dicts, indices, return_baskets=True)
+        return dataset, tensor_names, problematic_sample_ids, baskets
 
-    def _get_predictions(self, dataset, tensor_names, baskets, rest_api_schema=False):
+    def _get_predictions(self, dataset, tensor_names, baskets):
         """
         Feed a preprocessed dataset to the model and get the actual predictions (forward pass + formatting).
 
@@ -473,10 +565,6 @@ class Inferencer:
         :param baskets: For each item in the dataset, we need additional information to create formatted preds.
                         Baskets contain all relevant infos for that.
                         Example: QA - input string to convert the predicted answer from indices back to string space
-        :param rest_api_schema: Whether input dicts use the format that complies with the FARM REST API.
-                                Currently only used for QA to switch from squad to a more useful format in production.
-                                While input is almost the same, output contains additional meta data(offset, context..)
-        :type rest_api_schema: bool
         :return: list of predictions
         """
         samples = [s for b in baskets for s in b.samples]
@@ -496,13 +584,12 @@ class Inferencer:
                     logits=[logits],
                     samples=batch_samples,
                     tokenizer=self.processor.tokenizer,
-                    rest_api_schema=rest_api_schema,
                     return_class_probs=self.return_class_probs,
                     **batch)
                 preds_all += preds
         return preds_all
 
-    def _get_predictions_and_aggregate(self, dataset, tensor_names, baskets, rest_api_schema=False):
+    def _get_predictions_and_aggregate(self, dataset, tensor_names, baskets):
         """
         Feed a preprocessed dataset to the model and get the actual predictions (forward pass + logits_to_preds + formatted_preds).
 
@@ -515,18 +602,18 @@ class Inferencer:
         :param baskets: For each item in the dataset, we need additional information to create formatted preds.
                         Baskets contain all relevant infos for that.
                         Example: QA - input string to convert the predicted answer from indices back to string space
-        :param rest_api_schema: Whether input dicts use the format that complies with the FARM REST API.
-                                Currently only used for QA to switch from squad to a more useful format in production.
-                                While input is almost the same, output contains additional meta data(offset, context..)
-        :type rest_api_schema: bool
         :return: list of predictions
         """
 
         data_loader = NamedDataLoader(
             dataset=dataset, sampler=SequentialSampler(dataset), batch_size=self.batch_size, tensor_names=tensor_names
         )
+        # TODO Sometimes this is the preds of one head, sometimes of two. We need a more advanced stacking operation
+        # TODO so that preds of the right shape are passed in to formatted_preds
         unaggregated_preds_all = []
+
         for i, batch in enumerate(tqdm(data_loader, desc=f"Inferencing Samples", unit=" Batches", disable=self.disable_tqdm)):
+
             batch = {key: batch[key].to(self.device) for key in batch}
 
             # get logits
@@ -534,19 +621,24 @@ class Inferencer:
                 # Aggregation works on preds, not logits. We want as much processing happening in one batch + on GPU
                 # So we transform logits to preds here as well
                 logits = self.model.forward(**batch)
-                preds = self.model.logits_to_preds(logits, **batch)[0]
-                unaggregated_preds_all += preds
+                # preds = self.model.logits_to_preds(logits, **batch)[0] (This must somehow be useful for SQuAD)
+                preds = self.model.logits_to_preds(logits, **batch)
+                unaggregated_preds_all.append(preds)
 
         # In some use cases we want to aggregate the individual predictions.
         # This is mostly useful, if the input text is longer than the max_seq_len that the model can process.
         # In QA we can use this to get answers from long input texts by first getting predictions for smaller passages
         # and then aggregating them here.
 
+        # At this point unaggregated preds has shape [n_batches][n_heads][n_samples]
+
         # can assume that we have only complete docs i.e. all the samples of one doc are in the current chunk
-        preds_all = self.model.formatted_preds(logits=[None], # For QA we collected preds per batch and do not want to pass logits
-                                               preds_p=unaggregated_preds_all,
-                                               baskets=baskets,
-                                               rest_api_schema=rest_api_schema)[0]
+        logits = [None]
+        preds_all = self.model.formatted_preds(logits=logits, # For QA we collected preds per batch and do not want to pass logits
+                                               preds=unaggregated_preds_all,
+                                               baskets=baskets)
+        if self.benchmarking:
+            self.benchmarker.record("formatted_preds")
         return preds_all
 
     def extract_vectors(self, dicts, extraction_strategy="cls_token", extraction_layer=-1):
@@ -572,7 +664,47 @@ class Inferencer:
         self.model.language_model.extraction_layer = extraction_layer
         self.model.language_model.extraction_strategy = extraction_strategy
 
-        return self.inference_from_dicts(dicts, rest_api_schema=False)
+        return self.inference_from_dicts(dicts)
+
+
+class QAInferencer(Inferencer):
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        if self.task_type != "question_answering":
+            logger.warning("QAInferencer always has task_type='question_answering' even if another value is provided "
+                           "to Inferencer.load() or QAInferencer()")
+            self.task_type = "question_answering"
+
+    def inference_from_dicts(self,
+                             dicts,
+                             return_json=True,
+                             multiprocessing_chunksize=None,
+                             streaming=False) -> Union[List[QAPred], Generator[QAPred, None, None]]:
+        if isinstance(self.processor, NaturalQuestionsProcessor):
+            for questions_key in ['questions', 'qas']:
+                if questions_key in dicts[0].keys() and any([len(dict[questions_key]) > 1 for dict in dicts]):
+                    logger.warning('More than one question for document. NaturalQuestions inference will return just the answer to the first question.')
+        return Inferencer.inference_from_dicts(self, dicts, return_json=return_json,
+                                               multiprocessing_chunksize=multiprocessing_chunksize, streaming=streaming)
+
+    def inference_from_file(self,
+                            file,
+                            multiprocessing_chunksize=None,
+                            streaming=False,
+                            return_json=True) -> Union[List[QAPred], Generator[QAPred, None, None]]:
+        return Inferencer.inference_from_file(self, file, return_json=return_json,
+                                              multiprocessing_chunksize=multiprocessing_chunksize, streaming=streaming)
+
+    def inference_from_objects(self,
+                               objects: List[QAInput],
+                               return_json=True,
+                               multiprocessing_chunksize=None,
+                               streaming=False) -> Union[List[QAPred], Generator[QAPred, None, None]]:
+        dicts = [o.to_dict() for o in objects]
+        # TODO investigate this deprecation warning. Timo: I thought we were about to implement Input Objects, then we can and should use inference from (input) objects!
+        #logger.warning("QAInferencer.inference_from_objects() will soon be deprecated. Use QAInferencer.inference_from_dicts() instead")
+        return self.inference_from_dicts(dicts, return_json=return_json,
+                                         multiprocessing_chunksize=multiprocessing_chunksize, streaming=streaming)
 
 
 class FasttextInferencer:
