@@ -6,17 +6,17 @@ from pathlib import Path
 
 import multiprocessing
 import numpy
-import onnxruntime
 import torch
 from torch import nn
-from transformers.modeling_auto import AutoModelForQuestionAnswering, AutoModelForSequenceClassification, AutoModelForTokenClassification, AutoModelWithLMHead
+from transformers import AutoConfig
+from transformers.convert_graph_to_onnx import convert, quantize as quantize_model
 
-from farm.data_handler.data_silo import DataSilo
-from farm.data_handler.processor import SquadProcessor
+
+from farm.data_handler.processor import Processor
 from farm.modeling.language_model import LanguageModel
-from farm.modeling.prediction_head import PredictionHead, QuestionAnsweringHead, TokenClassificationHead, TextClassificationHead
-from farm.modeling.tokenization import Tokenizer
-from farm.utils import MLFlowLogger as MlLogger
+from farm.modeling.prediction_head import PredictionHead, pick_single_fn
+from farm.utils import MLFlowLogger as MlLogger, stack
+import farm.conversion.transformers as conv
 
 logger = logging.getLogger(__name__)
 
@@ -80,17 +80,55 @@ class BaseAdaptiveModel:
         :type kwargs: object
         :return: predictions in the right format
         """
-        all_preds = []
-        # collect preds from all heads
-        # TODO add switch between single vs multiple prediction heads
-        for head, logits_for_head in zip(
-            self.prediction_heads, logits
-        ):
-            preds = head.formatted_preds(
-                logits=logits_for_head, **kwargs
-            )
-            all_preds.append(preds)
-        return all_preds
+        n_heads = len(self.prediction_heads)
+
+        if n_heads == 0:
+            # just return LM output (e.g. useful for extracting embeddings at inference time)
+            preds_final = self.language_model.formatted_preds(logits=logits, **kwargs)
+
+        elif n_heads == 1:
+            preds_final = []
+            # This try catch is to deal with the fact that sometimes we collect preds before passing it to
+            # formatted_preds (see Inferencer._get_predictions_and_aggregate()) and sometimes we don't
+            # (see Inferencer._get_predictions())
+            try:
+                preds = kwargs["preds"]
+                temp = [y[0] for y in preds]
+                preds_flat = [item for sublist in temp for item in sublist]
+                kwargs["preds"] = preds_flat
+            except KeyError:
+                kwargs["preds"] = None
+            head = self.prediction_heads[0]
+            logits_for_head = logits[0]
+            preds = head.formatted_preds(logits=logits_for_head, **kwargs)
+            # TODO This is very messy - we need better definition of what the output should look like
+            if type(preds) == list:
+                preds_final += preds
+            elif type(preds) == dict and "predictions" in preds:
+                preds_final.append(preds)
+
+        # This case is triggered by Natural Questions
+        else:
+            preds_final = [list() for _ in range(n_heads)]
+            preds = kwargs["preds"]
+            preds_for_heads = stack(preds)
+            logits_for_heads = [None] * n_heads
+
+            samples = [s for b in kwargs["baskets"] for s in b.samples]
+            kwargs["samples"] = samples
+
+            del kwargs["preds"]
+
+            for i, (head, preds_for_head, logits_for_head) in enumerate(zip(self.prediction_heads, preds_for_heads, logits_for_heads)):
+                preds = head.formatted_preds(logits=logits_for_head, preds=preds_for_head, **kwargs)
+                preds_final[i].append(preds)
+
+            # Look for a merge() function amongst the heads and if a single one exists, apply it to preds_final
+            merge_fn = pick_single_fn(self.prediction_heads, "merge_formatted_preds")
+            if merge_fn:
+                preds_final = merge_fn(preds_final)
+
+        return preds_final
 
     def connect_heads_with_processor(self, tasks, require_labels=True):
         """
@@ -157,6 +195,12 @@ class BaseAdaptiveModel:
 
         return model_files, config_files
 
+def loss_per_head_sum(loss_per_head, global_step=None, batch=None):
+    """
+    Input: loss_per_head (list of tensors), global_step (int), batch (dict)
+    Output: aggregated loss (tensor)
+    """
+    return sum(loss_per_head)
 
 class AdaptiveModel(nn.Module, BaseAdaptiveModel):
     """ PyTorch implementation containing all the modelling needed for your NLP task. Combines a language
@@ -216,7 +260,7 @@ class AdaptiveModel(nn.Module, BaseAdaptiveModel):
         self.log_params()
         # default loss aggregation function is a simple sum (without using any of the optional params)
         if not loss_aggregation_fn:
-            loss_aggregation_fn = lambda loss_per_head, global_step=None, batch=None: sum(loss_per_head)
+            loss_aggregation_fn = loss_per_head_sum
         self.loss_aggregation_fn = loss_aggregation_fn
 
     def fit_heads_to_lm(self):
@@ -226,6 +270,22 @@ class AdaptiveModel(nn.Module, BaseAdaptiveModel):
             ph.resize_input(self.lm_output_dims)
             ph.to(self.device)
 
+    def bypass_ph(self):
+        """Replaces methods in the prediction heads with dummy functions. Used for benchmarking where we want to
+        isolate the lm run time from ph run time."""
+        def fake_forward(x):
+            """Slices lm vector outputs of shape (batch_size, max_seq_len, dims) --> (batch_size, max_seq_len, 2)"""
+            return x.narrow(2, 0, 2)
+        def fake_logits_to_preds(logits, **kwargs):
+            batch_size = logits.shape[0]
+            return [None, None] * batch_size
+        def fake_formatted_preds(**kwargs):
+            return None
+        for ph in self.prediction_heads:
+            ph.forward = fake_forward
+            ph.logits_to_preds = fake_logits_to_preds
+            ph.formatted_preds = fake_formatted_preds
+
     def save(self, save_dir):
         """
         Saves the language model and prediction heads. This will generate a config file
@@ -324,23 +384,6 @@ class AdaptiveModel(nn.Module, BaseAdaptiveModel):
         loss = self.loss_aggregation_fn(all_losses, global_step=global_step, batch=kwargs)
         return loss
 
-    def logits_to_preds(self, logits, **kwargs):
-        """
-        Get predictions from all prediction heads.
-
-        :param logits: logits, can vary in shape and type, depending on task
-        :type logits: object
-        :param label_maps: Maps from label encoding to label string
-        :param label_maps: dict
-        :return: A list of all predictions from all prediction heads
-        """
-        all_preds = []
-        # collect preds from all heads
-        for head, logits_for_head in zip(self.prediction_heads, logits):
-            preds = head.logits_to_preds(logits=logits_for_head, **kwargs)
-            all_preds.append(preds)
-        return all_preds
-
     def prepare_labels(self, **kwargs):
         """
         Label conversion to original label space, per prediction head.
@@ -358,29 +401,6 @@ class AdaptiveModel(nn.Module, BaseAdaptiveModel):
             all_labels.append(labels)
         return all_labels
 
-    def formatted_preds(self, logits, **kwargs):
-        """
-        Format predictions for inference.
-
-        :param logits: model logits
-        :type logits: torch.tensor
-        :param kwargs: placeholder for passing generic parameters
-        :type kwargs: object
-        :return: predictions in the right format
-        """
-        all_preds = []
-
-        if len(self.prediction_heads) == 0:
-            # just return LM output (e.g. useful for extracting embeddings at inference time)
-            all_preds = self.language_model.formatted_preds(logits=logits, **kwargs)
-        else:
-            # collect preds from all heads (default)
-            # TODO add switch between single vs multiple prediction heads
-            for head, logits_for_head in zip(self.prediction_heads, logits):
-                preds = head.formatted_preds(logits=logits_for_head, **kwargs)
-                all_preds.append(preds)
-        return all_preds
-
     def forward(self, **kwargs):
         """
         Push data through the whole model and returns logits. The data will propagate through the language
@@ -422,9 +442,11 @@ class AdaptiveModel(nn.Module, BaseAdaptiveModel):
     def forward_lm(self, **kwargs):
         """
         Forward pass for the language model
+
         :param kwargs:
         :return:
         """
+
         # Check if we have to extract from a special layer of the LM (default = last layer)
         try:
             extraction_layer = self.language_model.extraction_layer
@@ -433,11 +455,11 @@ class AdaptiveModel(nn.Module, BaseAdaptiveModel):
 
         # Run forward pass of language model
         if extraction_layer == -1:
-            sequence_output, pooled_output = self.language_model(**kwargs, output_all_encoded_layers=False)
+            sequence_output, pooled_output = self.language_model(**kwargs, return_dict=False, output_all_encoded_layers=False)
         else:
             # get output from an earlier layer
             self.language_model.enable_hidden_states_output()
-            sequence_output, pooled_output, all_hidden_states = self.language_model(**kwargs)
+            sequence_output, pooled_output, all_hidden_states = self.language_model(**kwargs, return_dict=False)
             sequence_output = all_hidden_states[extraction_layer]
             pooled_output = None #not available in earlier layers
             self.language_model.disable_hidden_states_output()
@@ -480,81 +502,16 @@ class AdaptiveModel(nn.Module, BaseAdaptiveModel):
         return self.language_model.language
 
     def convert_to_transformers(self):
-        if len(self.prediction_heads) != 1:
-            raise ValueError(f"Currently conversion only works for models with a SINGLE prediction head. "
-                             f"Your model has {len(self.prediction_heads)}")
-        elif len(self.prediction_heads[0].layer_dims) != 2:
-            raise ValueError(f"Currently conversion only works for PredictionHeads that are a single layer Feed Forward NN with dimensions [LM_output_dim, number_classes].\n"
-                             f"            Your PredictionHead has {str(self.prediction_heads[0].layer_dims)} dimensions.")
-        #TODO add more infos to config
+        """
+        Convert an adaptive model to huggingface's transformers format. Returns a list containing one model for each
+        prediction head.
 
-        if self.prediction_heads[0].model_type == "span_classification":
-            # init model
-            transformers_model = AutoModelForQuestionAnswering.from_config(self.language_model.model.config)
-            # transfer weights for language model + prediction head
-            setattr(transformers_model, transformers_model.base_model_prefix, self.language_model.model)
-            transformers_model.qa_outputs.load_state_dict(
-                self.prediction_heads[0].feed_forward.feed_forward[0].state_dict())
-
-        elif self.prediction_heads[0].model_type == "language_modelling":
-            # init model
-            transformers_model = AutoModelWithLMHead.from_config(self.language_model.model.config)
-            # transfer weights for language model + prediction head
-            setattr(transformers_model, transformers_model.base_model_prefix, self.language_model.model)
-            ph_state_dict = self.prediction_heads[0].state_dict()
-            ph_state_dict["transform.dense.weight"] = ph_state_dict.pop("dense.weight")
-            ph_state_dict["transform.dense.bias"] = ph_state_dict.pop("dense.bias")
-            ph_state_dict["transform.LayerNorm.weight"] = ph_state_dict.pop("LayerNorm.weight")
-            ph_state_dict["transform.LayerNorm.bias"] = ph_state_dict.pop("LayerNorm.bias")
-            transformers_model.cls.predictions.load_state_dict(ph_state_dict)
-            logger.warning("Currently only the Masked Language Modeling component of the prediction head is converted, "
-                           "not the Next Sentence Prediction or Sentence Order Prediction components")
-
-        elif self.prediction_heads[0].model_type == "text_classification":
-            if self.language_model.model.base_model_prefix == "roberta":
-                # Classification Heads in transformers have different architecture across Language Model variants
-                # The RobertaClassificationhead has components: input2dense, dropout, tanh, dense2output
-                # The tanh function cannot be mapped to current FARM style linear Feed Forward ClassificationHeads.
-                # So conversion for this type cannot work. We would need a compatible FARM RobertaClassificationHead
-                logger.error("Conversion for Text Classification with Roberta or XLMRoberta not possible at the moment.")
-                raise NotImplementedError
-
-            # add more info to config
-            self.language_model.model.config.id2label = {id: label for id, label in enumerate(self.prediction_heads[0].label_list)}
-            self.language_model.model.config.label2id = {label: id for id, label in enumerate(self.prediction_heads[0].label_list)}
-            self.language_model.model.config.finetuning_task = "text_classification"
-            self.language_model.model.config.language = self.language_model.language
-            self.language_model.model.config.num_labels = self.prediction_heads[0].num_labels
-
-            # init model
-            transformers_model = AutoModelForSequenceClassification.from_config(self.language_model.model.config)
-            # transfer weights for language model + prediction head
-            setattr(transformers_model, transformers_model.base_model_prefix, self.language_model.model)
-            transformers_model.classifier.load_state_dict(
-                self.prediction_heads[0].feed_forward.feed_forward[0].state_dict())
-        elif self.prediction_heads[0].model_type == "token_classification":
-            # add more info to config
-            self.language_model.model.config.id2label = {id: label for id, label in enumerate(self.prediction_heads[0].label_list)}
-            self.language_model.model.config.label2id = {label: id for id, label in enumerate(self.prediction_heads[0].label_list)}
-            self.language_model.model.config.finetuning_task = "token_classification"
-            self.language_model.model.config.language = self.language_model.language
-            self.language_model.model.config.num_labels = self.prediction_heads[0].num_labels
-
-            # init model
-            transformers_model = AutoModelForTokenClassification.from_config(self.language_model.model.config)
-            # transfer weights for language model + prediction head
-            setattr(transformers_model, transformers_model.base_model_prefix, self.language_model.model)
-            transformers_model.classifier.load_state_dict(
-                self.prediction_heads[0].feed_forward.feed_forward[0].state_dict())
-        else:
-            raise NotImplementedError(f"FARM -> Transformers conversion is not supported yet for"
-                                      f" prediction heads of type {self.prediction_heads[0].model_type}")
-        pass
-
-        return transformers_model
+        :return: List of huggingface transformers models.
+        """
+        return conv.Converter.convert_to_transformers(self)
 
     @classmethod
-    def convert_from_transformers(cls, model_name_or_path, device, task_type, processor=None):
+    def convert_from_transformers(cls, model_name_or_path, device, revision=None, task_type=None, processor=None):
         """
         Load a (downstream) model from huggingface's transformers format. Use cases:
          - continue training in FARM (e.g. take a squad QA model and fine-tune on your own data)
@@ -567,6 +524,8 @@ class AdaptiveModel(nn.Module, BaseAdaptiveModel):
                                               - deepset/bert-large-uncased-whole-word-masking-squad2
 
                                               See https://huggingface.co/models for full list
+        :param revision: The version of model to use from the HuggingFace model hub. Can be tag name, branch name, or commit hash.
+        :type revision: str
         :param device: "cpu" or "cuda"
         :param task_type: One of :
                           - 'question_answering'
@@ -577,138 +536,89 @@ class AdaptiveModel(nn.Module, BaseAdaptiveModel):
         :type processor: Processor
         :return: AdaptiveModel
         """
-        lm = LanguageModel.load(model_name_or_path)
-        #TODO Infer type of head automatically from config
+        return conv.Converter.convert_from_transformers(model_name_or_path,
+                                                        revision=revision,
+                                                        device=device,
+                                                        task_type=task_type,
+                                                        processor=processor)
 
-        if task_type == "question_answering":
-            ph = QuestionAnsweringHead.load(model_name_or_path)
-            adaptive_model = cls(language_model=lm, prediction_heads=[ph], embeds_dropout_prob=0.1,
-                               lm_output_types="per_token", device=device)
-        elif task_type == "text_classification":
-            if "roberta" in model_name_or_path:
-                # The RobertaClassificationhead has components: input2dense, dropout, tanh, dense2output
-                # The tanh function cannot be mapped to current FARM style linear Feed Forward PredictionHeads.
-                logger.error("Conversion for Text Classification with Roberta or XLMRoberta not possible at the moment.")
-                raise NotImplementedError
-            ph = TextClassificationHead.load(model_name_or_path)
-            adaptive_model = cls(language_model=lm, prediction_heads=[ph], embeds_dropout_prob=0.1,
-                                 lm_output_types="per_sequence", device=device)
-        elif task_type == "ner":
-            ph = TokenClassificationHead.load(model_name_or_path)
-            adaptive_model = cls(language_model=lm, prediction_heads=[ph], embeds_dropout_prob=0.1,
-                               lm_output_types="per_token", device=device)
-        elif task_type == "embeddings":
-            adaptive_model = cls(language_model=lm, prediction_heads=[], embeds_dropout_prob=0.1,
-                                 lm_output_types=["per_token", "per_sequence"], device=device)
-        else:
-            raise NotImplementedError(f"Huggingface's transformer models of type {task_type} are not supported yet")
 
-        if processor:
-            adaptive_model.connect_heads_with_processor(processor.tasks)
-
-        return adaptive_model
-
-    def convert_to_onnx(self, output_path, opset_version=11):
+    @classmethod
+    def convert_to_onnx(cls, model_name, output_path, task_type, convert_to_float16=False, quantize=False, opset_version=11):
         """
-        Convert a PyTorch AdaptiveModel to ONNX.
+        Convert a PyTorch model from transformers hub to an ONNX Model.
 
-        The conversion is trace-based by performing a forward pass on the model with a input batch.
-
-        :param output_path: model dir to write the model and config files
+        :param model_name: transformers model name
+        :type model_name: str
+        :param output_path: output Path to write the converted to
         :type output_path: Path
+        :param task_type: Type of task for the model. Available options: "embeddings", "question_answering",
+                          "text_classification", "ner".
+        :param convert_to_float16: By default, the model use float32 precision. With half precision of flaot16, inference
+                                should be faster on Nvidia GPUs with Tensor core like T4 or V100. On older GPUs, float32
+                                might be more performant.
+        :type convert_to_float16: bool
+        :param quantize: convert floating point number to integers
+        :type quantize: bool
         :param opset_version: ONNX opset version
         :type opset_version: int
         :return:
         """
-        if type(self.prediction_heads[0]) is not QuestionAnsweringHead:
-            raise NotImplementedError
+        language_model_class = LanguageModel.get_language_model_class(model_name)
+        if language_model_class not in ["Bert", "Roberta", "XLMRoberta"]:
+            raise Exception("The current ONNX conversion only support 'BERT', 'RoBERTa', and 'XLMRoberta' models.")
 
-        tokenizer = Tokenizer.load(
-            pretrained_model_name_or_path="deepset/bert-base-cased-squad2"
-        )
-
-        label_list = ["start_token", "end_token"]
-        metric = "squad"
-        max_seq_len = 384
-        batch_size = 1
-        processor = SquadProcessor(
-            tokenizer=tokenizer,
-            max_seq_len=max_seq_len,
-            label_list=label_list,
-            metric=metric,
-            train_filename="stub-file",  # the data is loaded from dicts instead of file.
-            dev_filename=None,
-            test_filename=None,
-            data_dir="stub-dir",
-        )
-
-        data_silo = DataSilo(processor=processor, batch_size=1, distributed=False, automatic_loading=False)
-        sample_dict = [
-            {
-                "context": 'The Normans were the people who in the 10th and 11th centuries gave their name to Normandy, '
-                           'a region in France. They were descended from Norse ("Norman" comes from "Norseman") raiders '
-                           'and pirates from Denmark, Iceland and Norway who, under their leader Rollo, agreed to swear '
-                           'fealty to King Charles III of West Francia.',
-                "qas": [
-                    {
-                        "question": "In what country is Normandy located?",
-                        "id": "56ddde6b9a695914005b9628",
-                        "answers": [{"text": "France", "answer_start": 159}],
-                        "is_impossible": False,
-                    }
-                ],
-            }
-        ]
-
-        data_silo._load_data(train_dicts=sample_dict)
-        data_loader = data_silo.get_data_loader("train")
-        data = next(iter(data_loader))
-        data = list(data.values())
-
-        inputs = {
-            'input_ids': data[0].to(self.device).reshape(batch_size, max_seq_len),
-            'padding_mask': data[1].to(self.device).reshape(batch_size, max_seq_len),
-            'segment_ids': data[2].to(self.device).reshape(batch_size, max_seq_len)
+        task_type_to_pipeline_map = {
+            "question_answering": "question-answering",
+            "embeddings": "feature-extraction",
+            "ner": "ner"
         }
 
-        # The method argument passing in torch.onnx.export is different to AdaptiveModel's forward().
-        # To resolve that, an ONNXWrapper instance is used.
-        model = ONNXWrapper.load_from_adaptive_model(self)
-
-        if not os.path.exists(output_path):
-            os.makedirs(output_path)
-
-        with torch.no_grad():
-            symbolic_names = {0: 'batch_size', 1: 'max_seq_len'}
-            torch.onnx.export(model,
-                              args=tuple(inputs.values()),
-                              f=output_path / 'model.onnx'.format(opset_version),
-                              opset_version=opset_version,
-                              do_constant_folding=True,
-                              input_names=['input_ids',
-                                           'padding_mask',
-                                           'segment_ids'],
-                              output_names=['logits'],
-                              dynamic_axes={'input_ids': symbolic_names,
-                                            'padding_mask': symbolic_names,
-                                            'segment_ids': symbolic_names,
-                                            'logits': symbolic_names,
-                                            })
-        # PredictionHead contains functionalities like logits_to_preds() that would still be needed
-        # for Inference with ONNX models. Only the config of the PredictionHead is stored.
-        for i, ph in enumerate(self.prediction_heads):
-            ph.save_config(output_path, i)
+        convert(
+            pipeline_name=task_type_to_pipeline_map[task_type],
+            framework="pt",
+            model=model_name,
+            output=output_path/"model.onnx",
+            opset=opset_version,
+            use_external_format=True if language_model_class is "XLMRoberta" else False
+        )
 
+        # save processor & model config files that are needed when loading the model with the FARM Inferencer
+        processor = Processor.convert_from_transformers(
+            tokenizer_name_or_path=model_name,
+            task_type=task_type,
+            max_seq_len=256,
+            doc_stride=128,
+            use_fast=True
+        )
         processor.save(output_path)
+        model = AdaptiveModel.convert_from_transformers(model_name, device="cpu", task_type=task_type)
+        model.save(output_path)
+        os.remove(output_path / "language_model.bin")  # remove the actual PyTorch model(only configs are required)
 
         onnx_model_config = {
+            "task_type": task_type,
             "onnx_opset_version": opset_version,
-            "language": self.get_language(),
+            "language_model_class": language_model_class,
+            "language": model.language_model.language
         }
-        with open(output_path / "model_config.json", "w") as f:
+        with open(output_path / "onnx_model_config.json", "w") as f:
             json.dump(onnx_model_config, f)
 
-        logger.info(f"Model exported at path {output_path}")
+        if convert_to_float16:
+            from onnxruntime_tools import optimizer
+            config = AutoConfig.from_pretrained(model_name)
+            optimized_model = optimizer.optimize_model(
+                input=str(output_path/"model.onnx"),
+                model_type='bert',
+                num_heads=config.num_hidden_layers,
+                hidden_size=config.hidden_size
+            )
+            optimized_model.convert_model_float32_to_float16()
+            optimized_model.save_model_to_file("model.onnx")
+
+        if quantize:
+            quantize_model(output_path/"model.onnx")
 
 
 class ONNXAdaptiveModel(BaseAdaptiveModel):
@@ -720,17 +630,20 @@ class ONNXAdaptiveModel(BaseAdaptiveModel):
 
     For inference, this class is compatible with the FARM Inferencer.
     """
-    def __init__(self, onnx_session, prediction_heads, language, device):
+    def __init__(self, onnx_session, language_model_class, language, prediction_heads, device):
         if str(device) == "cuda" and onnxruntime.get_device() != "GPU":
             raise Exception(f"Device {device} not available for Inference. For CPU, run pip install onnxruntime and"
                             f"for GPU run pip install onnxruntime-gpu")
         self.onnx_session = onnx_session
+        self.language_model_class = language_model_class
+        self.language = language
         self.prediction_heads = prediction_heads
         self.device = device
-        self.language = language
 
     @classmethod
     def load(cls, load_dir, device, **kwargs):
+        load_dir = Path(load_dir)
+        import onnxruntime
         sess_options = onnxruntime.SessionOptions()
         # Set graph optimization level to ORT_ENABLE_EXTENDED to enable bert optimization.
         sess_options.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_ENABLE_EXTENDED
@@ -749,11 +662,12 @@ class ONNXAdaptiveModel(BaseAdaptiveModel):
             prediction_heads.append(head)
             ph_output_type.append(head.ph_output_type)
 
-        with open(load_dir/"model_config.json") as f:
+        with open(load_dir/"onnx_model_config.json") as f:
             model_config = json.load(f)
+            language_model_class = model_config["language_model_class"]
             language = model_config["language"]
 
-        return cls(onnx_session, prediction_heads, language, device)
+        return cls(onnx_session, language_model_class, language, prediction_heads, device)
 
     def forward(self, **kwargs):
         """
@@ -763,13 +677,20 @@ class ONNXAdaptiveModel(BaseAdaptiveModel):
         :return: all logits as torch.tensor or multiple tensors.
         """
         with torch.no_grad():
-            input_to_onnx = {
-                'input_ids': numpy.ascontiguousarray(kwargs['input_ids'].cpu().numpy()),
-                'padding_mask': numpy.ascontiguousarray(kwargs['padding_mask'].cpu().numpy()),
-                'segment_ids': numpy.ascontiguousarray(kwargs['segment_ids'].cpu().numpy()),
-            }
+            if self.language_model_class == "Bert":
+                input_to_onnx = {
+                    'input_ids': numpy.ascontiguousarray(kwargs['input_ids'].cpu().numpy()),
+                    'attention_mask': numpy.ascontiguousarray(kwargs['padding_mask'].cpu().numpy()),
+                    'token_type_ids': numpy.ascontiguousarray(kwargs['segment_ids'].cpu().numpy()),
+                }
+            elif self.language_model_class in ["Roberta", "XLMRoberta"]:
+                input_to_onnx = {
+                    'input_ids': numpy.ascontiguousarray(kwargs['input_ids'].cpu().numpy()),
+                    'attention_mask': numpy.ascontiguousarray(kwargs['padding_mask'].cpu().numpy())
+                }
             res = self.onnx_session.run(None, input_to_onnx)
-            logits = [torch.from_numpy(res[0]).to(self.device)]
+            res = numpy.stack(res).transpose(1, 2, 0)
+            logits = [torch.Tensor(res).to(self.device)]
 
         return logits
 
