@@ -34,19 +34,25 @@ from torch import nn
 
 logger = logging.getLogger(__name__)
 
-from transformers.modeling_bert import BertModel, BertConfig
-from transformers.modeling_roberta import RobertaModel, RobertaConfig
-from transformers.modeling_xlnet import XLNetModel, XLNetConfig
-from transformers.modeling_albert import AlbertModel, AlbertConfig
-from transformers.modeling_xlm_roberta import XLMRobertaModel, XLMRobertaConfig
-from transformers.modeling_distilbert import DistilBertModel, DistilBertConfig
+from transformers import (
+    BertModel, BertConfig,
+    RobertaModel, RobertaConfig,
+    XLNetModel, XLNetConfig,
+    AlbertModel, AlbertConfig,
+    XLMRobertaModel, XLMRobertaConfig,
+    DistilBertModel, DistilBertConfig,
+    ElectraModel, ElectraConfig,
+    CamembertModel, CamembertConfig
+)
+
+from transformers import AutoModel, AutoConfig
 from transformers.modeling_utils import SequenceSummary
-from transformers.tokenization_bert import load_vocab
+from transformers.models.bert.tokenization_bert import load_vocab
+import transformers
 
 from farm.modeling import wordembedding_utils
 from farm.modeling.wordembedding_utils import s3e_pooling
 
-
 # These are the names of the attributes in various model configs which refer to the number of dimensions
 # in the output vectors
 OUTPUT_DIM_NAMES = ["dim", "hidden_size", "d_model"]
@@ -78,7 +84,7 @@ class LanguageModel(nn.Module):
         return model.from_scratch(vocab_size)
 
     @classmethod
-    def load(cls, pretrained_model_name_or_path, n_added_tokens=0, language_model_class=None, **kwargs):
+    def load(cls, pretrained_model_name_or_path, revision=None, n_added_tokens=0, language_model_class=None, **kwargs):
         """
         Load a pretrained language model either by
 
@@ -105,51 +111,46 @@ class LanguageModel(nn.Module):
         * albert-large-v2
         * distilbert-base-german-cased
         * distilbert-base-multilingual-cased
+        * google/electra-small-discriminator
+        * google/electra-base-discriminator
+        * google/electra-large-discriminator
+        * facebook/dpr-question_encoder-single-nq-base
+        * facebook/dpr-ctx_encoder-single-nq-base
 
         See all supported model variations here: https://huggingface.co/models
 
-        The appropriate language model class is inferred automatically from `pretrained_model_name_or_path`
+        The appropriate language model class is inferred automatically from model config
         or can be manually supplied via `language_model_class`.
 
         :param pretrained_model_name_or_path: The path of the saved pretrained model or its name.
         :type pretrained_model_name_or_path: str
+        :param revision: The version of model to use from the HuggingFace model hub. Can be tag name, branch name, or commit hash.
+        :type revision: str
         :param language_model_class: (Optional) Name of the language model class to load (e.g. `Bert`)
         :type language_model_class: str
 
         """
+        kwargs["revision"] = revision
+        logger.info("")
+        logger.info("LOADING MODEL")
+        logger.info("=============")
         config_file = Path(pretrained_model_name_or_path) / "language_model_config.json"
         if os.path.exists(config_file):
+            logger.info(f"Model found locally at {pretrained_model_name_or_path}")
             # it's a local directory in FARM format
             config = json.load(open(config_file))
             language_model = cls.subclasses[config["name"]].load(pretrained_model_name_or_path)
         else:
+            logger.info(f"Could not find {pretrained_model_name_or_path} locally.")
+            logger.info(f"Looking on Transformers Model Hub (in local cache and online)...")
             if language_model_class is None:
-                # it's transformers format (either from model hub or local)
-                pretrained_model_name_or_path = str(pretrained_model_name_or_path)
-                if "xlm" in pretrained_model_name_or_path and "roberta" in pretrained_model_name_or_path:
-                    language_model_class = 'XLMRoberta'
-                elif 'roberta' in pretrained_model_name_or_path:
-                    language_model_class = 'Roberta'
-                elif 'albert' in pretrained_model_name_or_path:
-                    language_model_class = 'Albert'
-                elif 'distilbert' in pretrained_model_name_or_path:
-                    language_model_class = 'DistilBert'
-                elif 'bert' in pretrained_model_name_or_path:
-                    language_model_class = 'Bert'
-                elif 'xlnet' in pretrained_model_name_or_path:
-                    language_model_class = 'XLNet'
-                elif "word2vec" in pretrained_model_name_or_path.lower() or "glove" in pretrained_model_name_or_path.lower():
-                    language_model_class = 'WordEmbedding_LM'
+                language_model_class = cls.get_language_model_class(pretrained_model_name_or_path)
 
             if language_model_class:
                 language_model = cls.subclasses[language_model_class].load(pretrained_model_name_or_path, **kwargs)
             else:
                 language_model = None
 
-            if language_model_class == 'XLMRoberta':
-                # TODO: for some reason, the pretrained XLMRoberta has different vocab size in the tokenizer compared to the model this is a hack to resolve that
-                n_added_tokens = 3
-
         if not language_model:
             raise Exception(
                 f"Model not found for {pretrained_model_name_or_path}. Either supply the local path for a saved "
@@ -158,6 +159,8 @@ class LanguageModel(nn.Module):
                 f"Transformers' model. Here's a list of available models: "
                 f"https://farm.deepset.ai/api/modeling.html#farm.modeling.language_model.LanguageModel.load"
             )
+        else:
+            logger.info(f"Loaded {pretrained_model_name_or_path}")
 
         # resize embeddings in case of custom vocab
         if n_added_tokens != 0:
@@ -173,6 +176,84 @@ class LanguageModel(nn.Module):
 
         return language_model
 
+    @staticmethod
+    def get_language_model_class(model_name_or_path):
+        # it's transformers format (either from model hub or local)
+        model_name_or_path = str(model_name_or_path)
+
+        config = AutoConfig.from_pretrained(model_name_or_path)
+        model_type = config.model_type
+        if model_type == "xlm-roberta":
+            language_model_class = "XLMRoberta"
+        elif model_type == "roberta":
+            if "mlm" in model_name_or_path.lower():
+                raise NotImplementedError("MLM part of codebert is currently not supported in FARM")
+            language_model_class = "Roberta"
+        elif model_type == "camembert":
+            language_model_class = "Camembert"
+        elif model_type == "albert":
+            language_model_class = "Albert"
+        elif model_type == "distilbert":
+            language_model_class = "DistilBert"
+        elif model_type == "bert":
+            language_model_class = "Bert"
+        elif model_type == "xlnet":
+            language_model_class = "XLNet"
+        elif model_type == "electra":
+            language_model_class = "Electra"
+        elif model_type == "dpr":
+            if config.architectures[0] == "DPRQuestionEncoder":
+                language_model_class = "DPRQuestionEncoder"
+            elif config.architectures[0] == "DPRContextEncoder":
+                language_model_class = "DPRContextEncoder"
+            elif config.archictectures[0] == "DPRReader":
+                raise NotImplementedError("DPRReader models are currently not supported.")
+        else:
+            # Fall back to inferring type from model name
+            logger.warning("Could not infer LanguageModel class from config. Trying to infer "
+                           "LanguageModel class from model name.")
+            language_model_class = LanguageModel._infer_language_model_class_from_string(model_name_or_path)
+
+        return language_model_class
+
+    @staticmethod
+    def _infer_language_model_class_from_string(model_name_or_path):
+        # If inferring Language model class from config doesn't succeed,
+        # fall back to inferring Language model class from model name.
+        if "xlm" in model_name_or_path.lower() and "roberta" in model_name_or_path.lower():
+            language_model_class = "XLMRoberta"
+        elif "roberta" in model_name_or_path.lower():
+            language_model_class = "Roberta"
+        elif "codebert" in model_name_or_path.lower():
+            if "mlm" in model_name_or_path.lower():
+                raise NotImplementedError("MLM part of codebert is currently not supported in FARM")
+            else:
+                language_model_class = "Roberta"
+        elif "camembert" in model_name_or_path.lower() or "umberto" in model_name_or_path.lower():
+            language_model_class = "Camembert"
+        elif "albert" in model_name_or_path.lower():
+            language_model_class = 'Albert'
+        elif "distilbert" in model_name_or_path.lower():
+            language_model_class = 'DistilBert'
+        elif "bert" in model_name_or_path.lower():
+            language_model_class = 'Bert'
+        elif "xlnet" in model_name_or_path.lower():
+            language_model_class = 'XLNet'
+        elif "electra" in model_name_or_path.lower():
+            language_model_class = 'Electra'
+        elif "word2vec" in model_name_or_path.lower() or "glove" in model_name_or_path.lower():
+            language_model_class = 'WordEmbedding_LM'
+        elif "minilm" in model_name_or_path.lower():
+            language_model_class = "Bert"
+        elif "dpr-question_encoder" in model_name_or_path.lower():
+            language_model_class = "DPRQuestionEncoder"
+        elif "dpr-ctx_encoder" in model_name_or_path.lower():
+            language_model_class = "DPRContextEncoder"
+        else:
+            language_model_class = None
+
+        return language_model_class
+
     def get_output_dims(self):
         config = self.model.config
         for odn in OUTPUT_DIM_NAMES:
@@ -232,19 +313,20 @@ class LanguageModel(nn.Module):
             "multilingual",
         )
         matches = [lang for lang in known_languages if lang in name]
-        if len(matches) == 0:
+        if "camembert" in name:
+            language = "french"
+            logger.info(
+                f"Automatically detected language from language model name: {language}"
+            )
+        elif "umberto" in name:
+            language = "italian"
+            logger.info(
+                f"Automatically detected language from language model name: {language}"
+            )
+        elif len(matches) == 0:
             language = "english"
-            logger.warning(
-                "Could not automatically detect from language model name what language it is. \n"
-                "\t We guess it's an *ENGLISH* model ... \n"
-                "\t If not: Init the language model by supplying the 'language' param."
-            )
         elif len(matches) > 1:
-            raise ValueError(
-                "Could not automatically detect from language model name what language it is.\n"
-                f"\t Found multiple matches: {matches}\n"
-                "\t Please init the language model by manually supplying the 'language' as a parameter.\n"
-            )
+            language = matches[0]
         else:
             language = matches[0]
             logger.info(
@@ -304,7 +386,7 @@ class LanguageModel(nn.Module):
         preds = []
         for vec, sample in zip(vecs, samples):
             pred = {}
-            pred["context"] = sample.tokenized["tokens"]
+            pred["context"] = sample.clear_text["text"]
             pred["vec"] = vec
             preds.append(pred)
         return preds
@@ -417,7 +499,7 @@ class Bert(LanguageModel):
             token_type_ids=segment_ids,
             attention_mask=padding_mask,
         )
-        if self.model.encoder.output_hidden_states == True:
+        if self.model.encoder.config.output_hidden_states == True:
             sequence_output, pooled_output, all_hidden_states = output_tuple[0], output_tuple[1], output_tuple[2]
             return sequence_output, pooled_output, all_hidden_states
         else:
@@ -425,10 +507,10 @@ class Bert(LanguageModel):
             return sequence_output, pooled_output
 
     def enable_hidden_states_output(self):
-        self.model.encoder.output_hidden_states = True
+        self.model.encoder.config.output_hidden_states = True
 
     def disable_hidden_states_output(self):
-        self.model.encoder.output_hidden_states = False
+        self.model.encoder.config.output_hidden_states = False
 
 
 class Albert(LanguageModel):
@@ -503,7 +585,7 @@ class Albert(LanguageModel):
             token_type_ids=segment_ids,
             attention_mask=padding_mask,
         )
-        if self.model.encoder.output_hidden_states == True:
+        if self.model.encoder.config.output_hidden_states == True:
             sequence_output, pooled_output, all_hidden_states = output_tuple[0], output_tuple[1], output_tuple[2]
             return sequence_output, pooled_output, all_hidden_states
         else:
@@ -511,10 +593,10 @@ class Albert(LanguageModel):
             return sequence_output, pooled_output
 
     def enable_hidden_states_output(self):
-        self.model.encoder.output_hidden_states = True
+        self.model.encoder.config.output_hidden_states = True
 
     def disable_hidden_states_output(self):
-        self.model.encoder.output_hidden_states = False
+        self.model.encoder.config.output_hidden_states = False
 
 
 class Roberta(LanguageModel):
@@ -590,7 +672,7 @@ class Roberta(LanguageModel):
             token_type_ids=segment_ids,
             attention_mask=padding_mask,
         )
-        if self.model.encoder.output_hidden_states == True:
+        if self.model.encoder.config.output_hidden_states == True:
             sequence_output, pooled_output, all_hidden_states = output_tuple[0], output_tuple[1], output_tuple[2]
             return sequence_output, pooled_output, all_hidden_states
         else:
@@ -598,10 +680,10 @@ class Roberta(LanguageModel):
             return sequence_output, pooled_output
 
     def enable_hidden_states_output(self):
-        self.model.encoder.output_hidden_states = True
+        self.model.encoder.config.output_hidden_states = True
 
     def disable_hidden_states_output(self):
-        self.model.encoder.output_hidden_states = False
+        self.model.encoder.config.output_hidden_states = False
 
 
 class XLMRoberta(LanguageModel):
@@ -677,7 +759,7 @@ class XLMRoberta(LanguageModel):
             token_type_ids=segment_ids,
             attention_mask=padding_mask,
         )
-        if self.model.encoder.output_hidden_states == True:
+        if self.model.encoder.config.output_hidden_states == True:
             sequence_output, pooled_output, all_hidden_states = output_tuple[0], output_tuple[1], output_tuple[2]
             return sequence_output, pooled_output, all_hidden_states
         else:
@@ -685,10 +767,10 @@ class XLMRoberta(LanguageModel):
             return sequence_output, pooled_output
 
     def enable_hidden_states_output(self):
-        self.model.encoder.output_hidden_states = True
+        self.model.encoder.config.output_hidden_states = True
 
     def disable_hidden_states_output(self):
-        self.model.encoder.output_hidden_states = False
+        self.model.encoder.config.output_hidden_states = False
 
 
 class DistilBert(LanguageModel):
@@ -734,7 +816,7 @@ class DistilBert(LanguageModel):
         farm_lm_config = Path(pretrained_model_name_or_path) / "language_model_config.json"
         if os.path.exists(farm_lm_config):
             # FARM style
-            config = AlbertConfig.from_pretrained(farm_lm_config)
+            config = DistilBertConfig.from_pretrained(farm_lm_config)
             farm_lm_model = Path(pretrained_model_name_or_path) / "language_model.bin"
             distilbert.model = DistilBertModel.from_pretrained(farm_lm_model, config=config, **kwargs)
             distilbert.language = distilbert.model.config.language
@@ -877,7 +959,7 @@ class XLNet(LanguageModel):
         )
         # XLNet also only returns the sequence_output (one vec per token)
         # We need to manually aggregate that to get a pooled output (one vec per seq)
-        #TODO verify that this is really doing correct pooling
+        # TODO verify that this is really doing correct pooling
         pooled_output = self.pooler(output_tuple[0])
 
         if self.model.output_hidden_states == True:
@@ -1096,8 +1178,11 @@ class WordEmbedding_LM(LanguageModel):
 
         sequence_output = torch.stack(sequence_output)
         pooled_output = torch.stack(pooled_output)
-        m = nn.BatchNorm1d(pooled_output.shape[1]) # batchnorm for stable learning
-        pooled_output = m(pooled_output)
+        m = nn.BatchNorm1d(pooled_output.shape[1])
+        # use batchnorm for more stable learning
+        # but disable it, if we have batch size of one (cannot compute batchnorm stats with only one sample)
+        if pooled_output.shape[0] > 1:
+            pooled_output = m(pooled_output)
         return sequence_output, pooled_output
 
     def trim_vocab(self, token_counts, processor, min_threshold):
@@ -1170,3 +1255,346 @@ class WordEmbedding_LM(LanguageModel):
                     ratio = (explained_variance[pca_idx] - explained_variance[PVN_dims]) / explained_variance[pca_idx]
                     self.model.embeddings[emb_idx] = self.model.embeddings[emb_idx] - ratio * np.dot(u.transpose(), self.model.embeddings[emb_idx]) * u
 
+
+class Electra(LanguageModel):
+    """
+    ELECTRA is a new pre-training approach which trains two transformer models:
+    the generator and the discriminator. The generator replaces tokens in a sequence,
+    and is therefore trained as a masked language model. The discriminator, which is
+    the model we're interested in, tries to identify which tokens were replaced by
+    the generator in the sequence.
+
+    The ELECTRA model here wraps HuggingFace's implementation
+    (https://github.com/huggingface/transformers) to fit the LanguageModel class.
+
+    NOTE:
+    - Electra does not output the pooled_output. An additional pooler is initialized.
+
+    """
+
+    def __init__(self):
+        super(Electra, self).__init__()
+        self.model = None
+        self.name = "electra"
+        self.pooler = None
+
+    @classmethod
+    def load(cls, pretrained_model_name_or_path, language=None, **kwargs):
+        """
+        Load a pretrained model by supplying
+
+        * the name of a remote model on s3 ("google/electra-base-discriminator" ...)
+        * OR a local path of a model trained via transformers ("some_dir/huggingface_model")
+        * OR a local path of a model trained via FARM ("some_dir/farm_model")
+
+        :param pretrained_model_name_or_path: The path of the saved pretrained model or its name.
+        :type pretrained_model_name_or_path: str
+
+        """
+
+        electra = cls()
+        if "farm_lm_name" in kwargs:
+            electra.name = kwargs["farm_lm_name"]
+        else:
+            electra.name = pretrained_model_name_or_path
+        # We need to differentiate between loading model using FARM format and Transformers format
+        farm_lm_config = Path(pretrained_model_name_or_path) / "language_model_config.json"
+        if os.path.exists(farm_lm_config):
+            # FARM style
+            config = ElectraConfig.from_pretrained(farm_lm_config)
+            farm_lm_model = Path(pretrained_model_name_or_path) / "language_model.bin"
+            electra.model = ElectraModel.from_pretrained(farm_lm_model, config=config, **kwargs)
+            electra.language = electra.model.config.language
+        else:
+            # Transformers Style
+            electra.model = ElectraModel.from_pretrained(str(pretrained_model_name_or_path), **kwargs)
+            electra.language = cls._get_or_infer_language_from_name(language, pretrained_model_name_or_path)
+        config = electra.model.config
+
+        # ELECTRA does not provide a pooled_output by default. Therefore, we need to initialize an extra pooler.
+        # The pooler takes the first hidden representation & feeds it to a dense layer of (hidden_dim x hidden_dim).
+        # We don't want a dropout in the end of the pooler, since we do that already in the adaptive model before we
+        # feed everything to the prediction head.
+        # Note: ELECTRA uses gelu as activation (BERT uses tanh instead)
+        config.summary_last_dropout = 0
+        config.summary_type = 'first'
+        config.summary_activation = 'gelu'
+        config.summary_use_proj = False
+        electra.pooler = SequenceSummary(config)
+        electra.pooler.apply(electra.model._init_weights)
+        return electra
+
+    def forward(
+        self,
+        input_ids,
+        segment_ids,
+        padding_mask,
+        **kwargs,
+    ):
+        """
+        Perform the forward pass of the ELECTRA model.
+
+        :param input_ids: The ids of each token in the input sequence. Is a tensor of shape [batch_size, max_seq_len]
+        :type input_ids: torch.Tensor
+        :param padding_mask: A mask that assigns a 1 to valid input tokens and 0 to padding tokens
+           of shape [batch_size, max_seq_len]
+        :return: Embeddings for each token in the input sequence.
+
+        """
+        output_tuple = self.model(
+            input_ids,
+            token_type_ids=segment_ids,
+            attention_mask=padding_mask,
+        )
+
+        # We need to manually aggregate that to get a pooled output (one vec per seq)
+        pooled_output = self.pooler(output_tuple[0])
+
+        if self.model.config.output_hidden_states == True:
+            sequence_output, all_hidden_states = output_tuple[0], output_tuple[1]
+            return sequence_output, pooled_output
+        else:
+            sequence_output = output_tuple[0]
+            return sequence_output, pooled_output
+
+    def enable_hidden_states_output(self):
+        self.model.config.output_hidden_states = True
+
+    def disable_hidden_states_output(self):
+        self.model.config.output_hidden_states = False
+
+
+class Camembert(Roberta):
+    """
+    A Camembert model that wraps the HuggingFace's implementation
+    (https://github.com/huggingface/transformers) to fit the LanguageModel class.
+    """
+    def __init__(self):
+        super(Camembert, self).__init__()
+        self.model = None
+        self.name = "camembert"
+
+    @classmethod
+    def load(cls, pretrained_model_name_or_path, language=None, **kwargs):
+        """
+        Load a language model either by supplying
+
+        * the name of a remote model on s3 ("camembert-base" ...)
+        * or a local path of a model trained via transformers ("some_dir/huggingface_model")
+        * or a local path of a model trained via FARM ("some_dir/farm_model")
+
+        :param pretrained_model_name_or_path: name or path of a model
+        :param language: (Optional) Name of language the model was trained for (e.g. "german").
+                         If not supplied, FARM will try to infer it from the model name.
+        :return: Language Model
+
+        """
+        camembert = cls()
+        if "farm_lm_name" in kwargs:
+            camembert.name = kwargs["farm_lm_name"]
+        else:
+            camembert.name = pretrained_model_name_or_path
+        # We need to differentiate between loading model using FARM format and Pytorch-Transformers format
+        farm_lm_config = Path(pretrained_model_name_or_path) / "language_model_config.json"
+        if os.path.exists(farm_lm_config):
+            # FARM style
+            config = CamembertConfig.from_pretrained(farm_lm_config)
+            farm_lm_model = Path(pretrained_model_name_or_path) / "language_model.bin"
+            camembert.model = CamembertModel.from_pretrained(farm_lm_model, config=config, **kwargs)
+            camembert.language = camembert.model.config.language
+        else:
+            # Huggingface transformer Style
+            camembert.model = CamembertModel.from_pretrained(str(pretrained_model_name_or_path), **kwargs)
+            camembert.language = cls._get_or_infer_language_from_name(language, pretrained_model_name_or_path)
+        return camembert
+
+
+class DPRQuestionEncoder(LanguageModel):
+    """
+    A DPRQuestionEncoder model that wraps HuggingFace's implementation
+    """
+
+    def __init__(self):
+        super(DPRQuestionEncoder, self).__init__()
+        self.model = None
+        self.name = "dpr_question_encoder"
+
+    @classmethod
+    def load(cls, pretrained_model_name_or_path, language=None, **kwargs):
+        """
+        Load a pretrained model by supplying
+
+        * the name of a remote model on s3 ("facebook/dpr-question_encoder-single-nq-base" ...)
+        * OR a local path of a model trained via transformers ("some_dir/huggingface_model")
+        * OR a local path of a model trained via FARM ("some_dir/farm_model")
+
+        :param pretrained_model_name_or_path: The path of the base pretrained language model whose weights are used to initialize DPRQuestionEncoder
+        :type pretrained_model_name_or_path: str
+        """
+
+        dpr_question_encoder = cls()
+        if "farm_lm_name" in kwargs:
+            dpr_question_encoder.name = kwargs["farm_lm_name"]
+        else:
+            dpr_question_encoder.name = pretrained_model_name_or_path
+
+        # We need to differentiate between loading model using FARM format and Pytorch-Transformers format
+        farm_lm_config = Path(pretrained_model_name_or_path) / "language_model_config.json"
+        if os.path.exists(farm_lm_config):
+            # FARM style
+            dpr_config = transformers.DPRConfig.from_pretrained(farm_lm_config)
+            farm_lm_model = Path(pretrained_model_name_or_path) / "language_model.bin"
+            dpr_question_encoder.model = transformers.DPRQuestionEncoder.from_pretrained(farm_lm_model, config=dpr_config, **kwargs)
+            dpr_question_encoder.language = dpr_question_encoder.model.config.language
+        else:
+            model_type = AutoConfig.from_pretrained(pretrained_model_name_or_path).model_type
+            if model_type == "dpr":
+                # "pretrained dpr model": load existing pretrained DPRQuestionEncoder model
+                dpr_question_encoder.model = transformers.DPRQuestionEncoder.from_pretrained(
+                    str(pretrained_model_name_or_path), **kwargs)
+            else:
+                # "from scratch": load weights from different architecture (e.g. bert) into DPRQuestionEncoder
+                dpr_question_encoder.model = transformers.DPRQuestionEncoder(config=transformers.DPRConfig(**kwargs))
+                dpr_question_encoder.model.base_model.bert_model = AutoModel.from_pretrained(
+                    str(pretrained_model_name_or_path), **kwargs)
+            dpr_question_encoder.language = cls._get_or_infer_language_from_name(language, pretrained_model_name_or_path)
+
+        return dpr_question_encoder
+
+    def forward(
+        self,
+        query_input_ids,
+        query_segment_ids,
+        query_attention_mask,
+        **kwargs,
+    ):
+        """
+        Perform the forward pass of the DPRQuestionEncoder model.
+
+        :param query_input_ids: The ids of each token in the input sequence. Is a tensor of shape [batch_size, max_seq_len]
+        :type query_input_ids: torch.Tensor
+        :param query_segment_ids: The id of the segment. For example, in next sentence prediction, the tokens in the
+           first sentence are marked with 0 and those in the second are marked with 1.
+           It is a tensor of shape [batch_size, max_seq_len]
+        :type query_segment_ids: torch.Tensor
+        :param query_attention_mask: A mask that assigns a 1 to valid input tokens and 0 to padding tokens
+           of shape [batch_size, max_seq_len]
+        :type query_attention_mask: torch.Tensor
+        :return: Embeddings for each token in the input sequence.
+
+        """
+        output_tuple = self.model(
+            input_ids=query_input_ids,
+            token_type_ids=query_segment_ids,
+            attention_mask=query_attention_mask,
+            return_dict=True
+        )
+        if self.model.question_encoder.config.output_hidden_states == True:
+            pooled_output, all_hidden_states = output_tuple.pooler_output, output_tuple.hidden_states
+            return pooled_output, all_hidden_states
+        else:
+            pooled_output = output_tuple.pooler_output
+            return pooled_output, None
+
+    def enable_hidden_states_output(self):
+        self.model.question_encoder.config.output_hidden_states = True
+
+    def disable_hidden_states_output(self):
+        self.model.question_encoder.config.output_hidden_states = False
+
+
+class DPRContextEncoder(LanguageModel):
+    """
+    A DPRContextEncoder model that wraps HuggingFace's implementation
+    """
+
+    def __init__(self):
+        super(DPRContextEncoder, self).__init__()
+        self.model = None
+        self.name = "dpr_context_encoder"
+
+    @classmethod
+    def load(cls, pretrained_model_name_or_path, language=None, **kwargs):
+        """
+        Load a pretrained model by supplying
+
+        * the name of a remote model on s3 ("facebook/dpr-ctx_encoder-single-nq-base" ...)
+        * OR a local path of a model trained via transformers ("some_dir/huggingface_model")
+        * OR a local path of a model trained via FARM ("some_dir/farm_model")
+
+        :param pretrained_model_name_or_path: The path of the base pretrained language model whose weights are used to initialize DPRContextEncoder
+        :type pretrained_model_name_or_path: str
+        """
+
+        dpr_context_encoder = cls()
+        if "farm_lm_name" in kwargs:
+            dpr_context_encoder.name = kwargs["farm_lm_name"]
+        else:
+            dpr_context_encoder.name = pretrained_model_name_or_path
+        # We need to differentiate between loading model using FARM format and Pytorch-Transformers format
+        farm_lm_config = Path(pretrained_model_name_or_path) / "language_model_config.json"
+        if os.path.exists(farm_lm_config):
+            # FARM style
+            dpr_config = transformers.DPRConfig.from_pretrained(farm_lm_config)
+            farm_lm_model = Path(pretrained_model_name_or_path) / "language_model.bin"
+            dpr_context_encoder.model = transformers.DPRContextEncoder.from_pretrained(farm_lm_model, config=dpr_config, **kwargs)
+            dpr_context_encoder.language = dpr_context_encoder.model.config.language
+        else:
+            # Pytorch-transformer Style
+            model_type = AutoConfig.from_pretrained(pretrained_model_name_or_path).model_type
+            if model_type == "dpr":
+                # "pretrained dpr model": load existing pretrained DPRContextEncoder model
+                dpr_context_encoder.model = transformers.DPRContextEncoder.from_pretrained(
+                    str(pretrained_model_name_or_path), **kwargs)
+            else:
+                # "from scratch": load weights from different architecture (e.g. bert) into DPRContextEncoder
+                dpr_context_encoder.model = transformers.DPRContextEncoder(config=transformers.DPRConfig(**kwargs))
+                dpr_context_encoder.model.base_model.bert_model = AutoModel.from_pretrained(
+                    str(pretrained_model_name_or_path), **kwargs)
+            dpr_context_encoder.language = cls._get_or_infer_language_from_name(language, pretrained_model_name_or_path)
+
+        return dpr_context_encoder
+
+    def forward(
+        self,
+        passage_input_ids,
+        passage_segment_ids,
+        passage_attention_mask,
+        **kwargs,
+    ):
+        """
+        Perform the forward pass of the DPRContextEncoder model.
+
+        :param passage_input_ids: The ids of each token in the input sequence. Is a tensor of shape [batch_size, number_of_hard_negative_passages, max_seq_len]
+        :type passage_input_ids: torch.Tensor
+        :param passage_segment_ids: The id of the segment. For example, in next sentence prediction, the tokens in the
+           first sentence are marked with 0 and those in the second are marked with 1.
+           It is a tensor of shape [batch_size, number_of_hard_negative_passages, max_seq_len]
+        :type passage_segment_ids: torch.Tensor
+        :param passage_attention_mask: A mask that assigns a 1 to valid input tokens and 0 to padding tokens
+           of shape [batch_size,  number_of_hard_negative_passages, max_seq_len]
+        :return: Embeddings for each token in the input sequence.
+
+        """
+        max_seq_len = passage_input_ids.shape[-1]
+        passage_input_ids = passage_input_ids.view(-1, max_seq_len)
+        passage_segment_ids = passage_segment_ids.view(-1, max_seq_len)
+        passage_attention_mask = passage_attention_mask.view(-1, max_seq_len)
+        output_tuple = self.model(
+            input_ids=passage_input_ids,
+            token_type_ids=passage_segment_ids,
+            attention_mask=passage_attention_mask,
+            return_dict=True
+        )
+        if self.model.ctx_encoder.config.output_hidden_states == True:
+            pooled_output, all_hidden_states = output_tuple.pooler_output, output_tuple.hidden_states
+            return pooled_output, all_hidden_states
+        else:
+            pooled_output = output_tuple.pooler_output
+            return pooled_output, None
+
+    def enable_hidden_states_output(self):
+        self.model.ctx_encoder.config.output_hidden_states = True
+
+    def disable_hidden_states_output(self):
+        self.model.ctx_encoder.config.output_hidden_states = False
